from __future__ import annotations

from graphql.type.definition import GraphQLList, GraphQLNonNull
from graphql.type.scalars import GraphQLBoolean, GraphQLFloat, GraphQLID, GraphQLInt, GraphQLString

from ._definition import ScalarType
from .any import GraphQLAny
from .base16 import GraphQLBase16
from .base32 import GraphQLBase32
from .base64 import GraphQLBase64
from .date import GraphQLDate
from .datetime import GraphQLDateTime
from .decimal import GraphQLDecimal
from .duration import GraphQLDuration
from .email import GraphQLEmail
from .file import GraphQLFile
from .image import GraphQLImage
from .ip import GraphQLIP
from .ipv4 import GraphQLIPv4
from .ipv6 import GraphQLIPv6
from .json import GraphQLJSON
from .null import GraphQLNull
from .time import GraphQLTime
from .url import GraphQLURL
from .uuid import GraphQLUUID

__all__ = [
    "GraphQLAny",
    "GraphQLBase16",
    "GraphQLBase32",
    "GraphQLBase64",
    "GraphQLBoolean",
    "GraphQLDate",
    "GraphQLDateTime",
    "GraphQLDecimal",
    "GraphQLDuration",
    "GraphQLEmail",
    "GraphQLFile",
    "GraphQLFloat",
    "GraphQLID",
    "GraphQLIP",
    "GraphQLIPv4",
    "GraphQLIPv6",
    "GraphQLImage",
    "GraphQLInt",
    "GraphQLJSON",
    "GraphQLList",
    "GraphQLNonNull",
    "GraphQLNull",
    "GraphQLString",
    "GraphQLTime",
    "GraphQLURL",
    "GraphQLUUID",
    "ScalarType",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphql import FieldNode, FragmentSpreadNode, GraphQLError, GraphQLObjectType, InlineFragmentNode, ValidationRule

from undine.settings import undine_settings
from undine.utils.graphql.undine_extensions import get_undine_entrypoint, get_undine_field
from undine.utils.graphql.utils import get_underlying_type

if TYPE_CHECKING:
    from graphql import (
        GraphQLAbstractType,
        GraphQLCompositeType,
        OperationDefinitionNode,
        SelectionNode,
        ValidationContext,
        VisitorAction,
    )


__all__ = [
    "MaxComplexityRule",
]


class MaxComplexityRule(ValidationRule):
    """Validates that the complexity of a GraphQL query does not exceed the maximum allowed."""

    def __init__(self, context: ValidationContext) -> None:
        super().__init__(context)
        self.complexity: int = 0
        self.visited_fragments: set[str] = set()

    def enter_operation_definition(self, node: OperationDefinitionNode, *_args: Any) -> VisitorAction:
        root_type = self.context.get_type()
        if not isinstance(root_type, GraphQLObjectType):
            return self.IDLE

        for selection in node.selection_set.selections:
            self.visited_fragments = set()
            self.handle_selection(root_type, selection)

        if self.complexity > undine_settings.MAX_QUERY_COMPLEXITY:
            msg = (
                f"Query complexity of {self.complexity} exceeds the maximum allowed "
                f"complexity of {undine_settings.MAX_QUERY_COMPLEXITY}."
            )
            error = GraphQLError(msg, node)
            self.context.report_error(error)
            return self.BREAK

        return self.IDLE

    def handle_selection(self, parent_type: GraphQLCompositeType, selection: SelectionNode) -> None:
        match selection:
            case FieldNode():
                self.handle_field(parent_type, selection)

            case FragmentSpreadNode():
                self.handle_fragment_spread(parent_type, selection)

            case InlineFragmentNode():
                self.handle_inline_fragment(parent_type, selection)  # type: ignore[arg-type]

    def handle_field(self, parent_type: GraphQLCompositeType, field_node: FieldNode) -> None:
        # Ignore fields on interfaces, as well as union '__typename'.
        if not isinstance(parent_type, GraphQLObjectType):
            return

        graphql_field = parent_type.fields.get(field_node.name.value)
        if graphql_field is None:
            return

        undine_entrypoint = get_undine_entrypoint(graphql_field)
        if undine_entrypoint is not None:
            self.complexity += undine_entrypoint.complexity

        undine_field = get_undine_field(graphql_field)
        if undine_field is not None:
            self.complexity += undine_field.complexity

        if field_node.selection_set is not None:
            field_type: GraphQLObjectType = get_underlying_type(graphql_field.type)

            for selection in field_node.selection_set.selections:
                self.handle_selection(field_type, selection)

    def handle_fragment_spread(self, parent_type: GraphQLCompositeType, fragment_spread: FragmentSpreadNode) -> None:
        fragment_name = fragment_spread.name.value
        if fragment_name in self.visited_fragments:
            return

        self.visited_fragments.add(fragment_name)

        fragment = self.context.get_fragment(fragment_name)
        if fragment is None:
            return

        for selection in fragment.selection_set.selections:
            self.handle_selection(parent_type, selection)

    def handle_inline_fragment(self, parent_type: GraphQLAbstractType, inline_fragment: InlineFragmentNode) -> None:
        fragment_type_name = inline_fragment.type_condition.name.value
        fragment_type: GraphQLObjectType = self.context.schema.get_type(fragment_type_name)  # type: ignore[assignment]

        for selection in inline_fragment.selection_set.selections:
            self.handle_selection(fragment_type, selection)



from __future__ import annotations

import dataclasses
import os
import time
import traceback
from contextlib import contextmanager, suppress
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any

import sqlparse
from django import db
from django.conf import settings

from undine.settings import undine_settings

if TYPE_CHECKING:
    from collections.abc import Callable, Generator

__all__ = [
    "capture_database_queries",
]


@dataclasses.dataclass(kw_only=True)
class QueryInfo:
    sql: str
    duration_ns: int
    origin: str


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class DBQueryData:
    queries: list[QueryInfo]
    enabled: bool = True

    @property
    def count(self) -> int:
        return len(self.queries)

    @property
    def log(self) -> str:
        message = "\n" + "-" * 75
        message += f"\n\n>>> Queries: ({len(self.queries)})"

        for index, info in enumerate(self.queries):
            message += "\n\n"
            message += f"{index + 1}) Duration: {info.duration_ns / 1_000_000:.2f} ms"
            message += "\n\n"
            message += "--- Query ".ljust(75, "-")
            message += "\n\n"
            message += sqlparse.format(info.sql, reindent=True)
            message += "\n\n"
            message += "--- Point of origin ".ljust(75, "-")
            message += "\n\n"
            message += info.origin
            message += "\n"
            message += "-" * 75

        return message


def db_query_logger(  # noqa: PLR0917
    execute: Callable[..., Any],
    sql: str,
    params: tuple[Any, ...],
    many: bool,  # noqa: FBT001
    context: dict[str, Any],
    # Added with functools.partial()
    query_data: DBQueryData,
) -> Any:
    """
    A database query logger for capturing executed database queries.
    Used to check that query optimizations work as expected.

    Can also be used as a place to put debugger breakpoint for solving issues.
    """
    # Don't include transaction creation, as we aren't interested in them.
    if sql.startswith(("SAVEPOINT", "RELEASE SAVEPOINT")):
        return execute(sql, params, many, context)

    sql_fmt = sql
    with suppress(TypeError):
        sql_fmt %= params

    info = QueryInfo(sql=sql_fmt, duration_ns=0, origin=get_stack_info())
    query_data.queries.append(info)

    start = time.perf_counter_ns()
    try:
        result = execute(sql, params, many, context)
    finally:
        info.duration_ns = time.perf_counter_ns() - start

    return result


def get_stack_info() -> str:
    if undine_settings.TESTING_CLIENT_FULL_STACKTRACE:
        return "".join(traceback.StackSummary.from_list(traceback.extract_stack()).format())

    venv_path: str | None = os.getenv("VIRTUAL_ENV")
    undine_path = str(Path(__file__).resolve().parent.parent)
    project_path = str(settings.BASE_DIR)

    for frame in reversed(traceback.extract_stack()):
        if frame.filename == __file__:
            continue

        if venv_path is not None and frame.filename.startswith(venv_path):
            continue

        if frame.filename.startswith(undine_path):
            return "".join(traceback.StackSummary.from_list([frame]).format())

        if frame.filename.startswith(project_path):
            return "".join(traceback.StackSummary.from_list([frame]).format())

    return "No info"


@contextmanager
def capture_database_queries(*, enabled: bool = True) -> Generator[DBQueryData, None, None]:
    """
    Capture results of what database queries were executed.

    :param enabled: Can be used to disable capturing.
    """
    query_data = DBQueryData(queries=[], enabled=enabled)
    if not enabled:
        yield query_data
        return

    query_logger = partial(db_query_logger, query_data=query_data)

    with db.connection.execute_wrapper(query_logger):
        yield query_data



from __future__ import annotations

import asyncio
import dataclasses
from asyncio import gather, get_running_loop
from typing import TYPE_CHECKING, Any, Generic, Self, TypeVar

from django.core import signals

from undine.exceptions import (
    GraphQLDataLoaderDidNotReturnSortedSequenceError,
    GraphQLDataLoaderPrimingError,
    GraphQLDataLoaderWrongNumberOfValuesReturnedError,
)

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop, Future
    from collections.abc import Callable, Coroutine, Hashable, Iterable, MutableMapping

    from undine.typing import SortedSequence


__all__ = [
    "DataLoader",
]


TKey = TypeVar("TKey")
TResult = TypeVar("TResult")


class DataLoader(Generic[TKey, TResult]):
    """A utility for loading data in batches. Requires an async server."""

    def __init__(
        self,
        *,
        load_fn: Callable[[list[TKey]], Coroutine[None, None, SortedSequence[TResult | BaseException]]],
        max_batch_size: int | None = None,
        reuse_loads: bool = True,
        key_hash_fn: Callable[[TKey], Hashable] = lambda x: x,
        lock: asyncio.Lock | None = None,
    ) -> None:
        """
        Create a new DataLoader.

        :param load_fn: Coroutine function used to load the data for the given keys.
        :param max_batch_size: Maximum number of keys to load in a single batch.
                                If `None`, all keys will be loaded in a single batch.
        :param reuse_loads: Whether loads should be reused for the same load key.
        :param key_hash_fn: Function used to generate a hash from a load key.
                            Loads with the same hash will be reused if `reuse_loads` is `True`.
                            Required if load key is not hashable.
        :param lock: A lock used to synchronize load function execution.
                     Can provide a common lock for multiple DataLoaders
                     if they can load the same object from different keys.
                     This way they don't fetch the same object twice.
        """
        self.load_fn = load_fn
        self.max_batch_size = max_batch_size
        self.reuse_loads = reuse_loads
        self.key_hash_fn = key_hash_fn
        self.lock = lock or asyncio.Lock()

        self.reusable_loads: MutableMapping[Hashable, DataLoaderFuture[TKey, TResult]] = {}
        """Loads that have been scheduled or completed and can be reused for the given key."""

        self.current_batch: DataLoaderBatch[TKey, TResult] = DataLoaderBatch(loader=self, dispatched=True)
        """Current batch of loads."""

        # Loads are only reused during the same request.
        # For reusing across requests or web server workers, you should implement caching yourself.
        # This should likely be done in the `load_fn` function.
        signals.request_finished.connect(self._request_finished)

    @property
    def loop(self) -> AbstractEventLoop:
        return get_running_loop()

    @property
    def should_create_new_batch(self) -> bool:
        current_batch = self.current_batch
        return current_batch.dispatched or (self.max_batch_size and len(current_batch.loads) >= self.max_batch_size)

    def load(self, key: TKey) -> Future[TResult | BaseException]:
        """Schedule a load for the given key."""
        if self.reuse_loads:
            load = self.reusable_loads.get(self.key_hash_fn(key))
            if load is not None and not load.future.cancelled():
                return load.future

        if self.should_create_new_batch:
            self.current_batch = DataLoaderBatch(loader=self)

        load = DataLoaderFuture(key=key, future=self.loop.create_future())

        self.current_batch.loads.append(load)
        if self.reuse_loads:
            self.reusable_loads[self.key_hash_fn(key)] = load

        return load.future

    def load_many(self, keys: Iterable[TKey]) -> Future[list[TResult | BaseException]]:
        """Schedule loads for the given keys."""
        return gather(*map(self.load, keys), return_exceptions=True)

    def clear(self, key: TKey) -> Self:
        """Remove a load by the given key from the reusable loads."""
        if self.reuse_loads:
            self.reusable_loads.pop(self.key_hash_fn(key), None)
        return self

    def clear_many(self, keys: Iterable[TKey]) -> Self:
        """Remove loads by the given keys from the reusable loads."""
        if self.reuse_loads:
            for key in keys:
                self.reusable_loads.pop(self.key_hash_fn(key), None)
        return self

    def clear_all(self) -> Self:
        """Remove all loads from the reusable loads."""
        if self.reuse_loads:
            self.reusable_loads.clear()
        return self

    def prime(
        self,
        key: TKey,
        value: TResult | BaseException,
        *,
        can_prime_pending_loads: bool = False,
    ) -> Self:
        """
        Add a value to reusable loads for the given key.

        :param key: The key to prime.
        :param value: The value to prime.
        :param can_prime_pending_loads: If `True`, pending loads for the key can also be primed.
        """
        return self.prime_many(keys=[key], values=[value], can_prime_pending_loads=can_prime_pending_loads)

    def prime_many(
        self,
        keys: SortedSequence[TKey],
        values: SortedSequence[TResult | BaseException],
        *,
        can_prime_pending_loads: bool = False,
    ) -> Self:
        """
        Add values to reusable loads for the given keys.
        A key in the keys sequence should match the value at the same index in the values sequence.

        :param keys: The keys to prime.
        :param values: The values to prime.
        :param can_prime_pending_loads: If `True`, pending loads for the key can also be primed.
        """
        if not self.reuse_loads:
            return self

        if len(keys) != len(values):
            raise GraphQLDataLoaderPrimingError(keys=len(keys), values=len(values))

        loads_changed = False

        for key, value in zip(keys, values, strict=False):
            key_hash = self.key_hash_fn(key)
            if key_hash in self.reusable_loads:
                if can_prime_pending_loads:
                    load = self.reusable_loads[key_hash]
                    if not load.future.done():
                        # Some loads in the current batch might now be completed and can be removed from it.
                        loads_changed = True
                        if isinstance(value, BaseException):
                            load.future.set_exception(value)
                        else:
                            load.future.set_result(value)

                continue

            # There might be loads for the same keys in the current batch
            # if a load was first cleared and then primed again.
            loads_changed = True
            load = DataLoaderFuture(key=key, future=self.loop.create_future())
            self.reusable_loads[key_hash] = load

            if isinstance(value, BaseException):
                load.future.set_exception(value)
            else:
                load.future.set_result(value)

        if loads_changed:
            self._prime_batch(keys, values)

        return self

    def _prime_batch(self, keys: SortedSequence[TKey], values: SortedSequence[TResult | BaseException]) -> None:
        """Set values for loads in the current batch for the given keys."""
        if self.current_batch.dispatched:
            return

        if len(keys) != len(values):
            raise GraphQLDataLoaderPrimingError(keys=len(keys), values=len(values))

        batch_updated = False

        for load in self.current_batch.loads:
            if load.future.done():
                continue

            if load.key in keys:
                batch_updated = True
                value = values[keys.index(load.key)]

                if isinstance(value, BaseException):
                    load.future.set_exception(value)
                else:
                    load.future.set_result(value)

        if batch_updated:
            self.current_batch.loads[:] = [load for load in self.current_batch.loads if not load.future.done()]

    def _request_finished(self, sender: type, **kwargs: Any) -> None:
        """Hook to run after each request to do some cleanup."""
        # Clear reusable loads so that we don't reuse them for the next request.
        self.clear_all()
        # Create a new empty dispatched batch so that completed loads from current batch can be freed from memory.
        self.current_batch = DataLoaderBatch(loader=self, dispatched=True)


@dataclasses.dataclass(slots=True, kw_only=True)
class DataLoaderBatch(Generic[TKey, TResult]):
    """A batch of loads to be loaded for a DataLoader."""

    loader: DataLoader[TKey, TResult]
    """The DataLoader this batch belongs to."""

    loads: list[DataLoaderFuture[TKey, TResult]] = dataclasses.field(default_factory=list)
    """Loads that have been scheduled in this batch."""

    dispatched: bool = False
    """Whether the batch has started loading data or not."""

    def __post_init__(self) -> None:
        if not self.dispatched:
            # Schedule the batch to be dispatched after all resolvers in the operation have been executed.
            # This also retains the reference to the batch if a new batch is created.
            self.loader.loop.create_task(self.dispatch())

    async def dispatch(self) -> None:
        """Execute the load function and set the results for the loads in the batch."""
        async with self.loader.lock:
            self.dispatched = True

            self.loads[:] = [load for load in self.loads if not load.future.done()]
            if not self.loads:
                return

            keys = [load.key for load in self.loads]

            try:
                values = await self.load(keys)

                if not isinstance(values, list | tuple):
                    raise GraphQLDataLoaderDidNotReturnSortedSequenceError(got=type(values))  # noqa: TRY301

                if len(values) != len(keys):
                    raise GraphQLDataLoaderWrongNumberOfValuesReturnedError(got=len(values), expected=len(keys))  # noqa: TRY301

            except Exception as error:  # noqa: BLE001
                for load in self.loads:
                    if not load.future.done():
                        load.future.set_exception(error)
                return

            for load, value in zip(self.loads, values, strict=True):
                if load.future.done():
                    continue

                if isinstance(value, BaseException):
                    load.future.set_exception(value)
                else:
                    load.future.set_result(value)

    def load(self, keys: list[TKey]) -> asyncio.Task[list[TResult]]:
        """Schedule the load function to run with the given keys."""
        load_function_task = self.loader.loop.create_task(self.loader.load_fn(keys))

        # Cancel the load function task in case a future expecting results from it is canceled
        # and all other futures are done while the load is still running. This can happen, for example,
        # when a TaskGroup is canceled due to an exception in one of its tasks. Otherwise,
        # the load function might use resources that are no longer available, e.g. a database connection.
        def callback(future: Future[TResult]) -> None:
            if future.cancelled() and all(load_.future.done() for load_ in self.loads):
                load_function_task.cancel()

        for load in self.loads:
            if not load.future.done():
                load.future.add_done_callback(callback)

        return load_function_task


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class DataLoaderFuture(Generic[TKey, TResult]):
    """A Future in a DataLoaderBatch where the data is loaded."""

    key: TKey
    """The load key for this load."""

    future: Future[TResult]
    """The future that will be set when the load is completed."""



from __future__ import annotations

from collections.abc import Hashable
from contextlib import contextmanager
from http import HTTPStatus
from typing import TYPE_CHECKING, Any, TypeGuard, TypeVar

from django.db.models import ForeignKey
from graphql import (
    DocumentNode,
    ExecutionResult,
    FieldNode,
    GraphQLEnumType,
    GraphQLError,
    GraphQLIncludeDirective,
    GraphQLInputObjectType,
    GraphQLInterfaceType,
    GraphQLObjectType,
    GraphQLResolveInfo,
    GraphQLScalarType,
    GraphQLSkipDirective,
    GraphQLUnionType,
    OperationDefinitionNode,
    OperationType,
    Undefined,
    get_argument_values,
    get_directive_values,
)

from undine.exceptions import (
    DirectiveLocationError,
    GraphQLErrorGroup,
    GraphQLRequestMultipleOperationsNoOperationNameError,
    GraphQLRequestNoOperationError,
    GraphQLRequestOperationNotFoundError,
)
from undine.settings import undine_settings
from undine.utils.logging import log_traceback
from undine.utils.model_utils import get_validation_error_messages
from undine.utils.reflection import get_traceback
from undine.utils.text import to_snake_case

if TYPE_CHECKING:
    from collections.abc import Collection, Generator, Iterable

    from django.core.exceptions import ValidationError
    from graphql import (
        DirectiveLocation,
        DocumentNode,
        GraphQLCompositeType,
        GraphQLField,
        GraphQLList,
        GraphQLNonNull,
        GraphQLOutputType,
        GraphQLSchema,
        GraphQLWrappingType,
        Node,
        SelectionNode,
    )
    from graphql.execution.values import NodeWithDirective

    from undine import Field, GQLInfo
    from undine.directives import Directive
    from undine.typing import ModelField


__all__ = [
    "check_directives",
    "get_arguments",
    "get_error_execution_result",
    "get_error_execution_result",
    "get_operation_definition",
    "get_operation_type",
    "get_queried_field_name",
    "get_underlying_type",
    "graphql_error_path",
    "graphql_errors_hook",
    "is_atomic_mutation",
    "is_connection",
    "is_edge",
    "is_node_interface",
    "is_page_info",
    "is_relation_id",
    "should_skip_node",
]


TGraphQLType = TypeVar(
    "TGraphQLType",
    GraphQLScalarType,
    GraphQLObjectType,
    GraphQLInterfaceType,
    GraphQLUnionType,
    GraphQLEnumType,
    GraphQLInputObjectType,
)


# Getters


def get_underlying_type(
    gql_type: (
        TGraphQLType
        | GraphQLList[TGraphQLType]
        | GraphQLList[GraphQLNonNull[TGraphQLType]]
        | GraphQLNonNull[TGraphQLType]
        | GraphQLNonNull[GraphQLList[TGraphQLType]]
        | GraphQLNonNull[GraphQLList[GraphQLNonNull[TGraphQLType]]]
        | GraphQLWrappingType[TGraphQLType]
    ),
) -> TGraphQLType:
    while hasattr(gql_type, "of_type"):
        gql_type = gql_type.of_type
    return gql_type


def get_arguments(info: GQLInfo) -> dict[str, Any]:
    """Get input arguments for the current field from the GraphQL resolve info."""
    graphql_field = info.parent_type.fields[info.field_name]
    return get_argument_values(graphql_field, info.field_nodes[0], info.variable_values)


def get_queried_field_name(original_name: str, info: GQLInfo) -> str:
    """Get the name of a field in the current query."""
    return original_name if info.path.key == info.field_name else info.path.key  # type: ignore[return-value]


def get_field_def(schema: GraphQLSchema, parent_type: GraphQLCompositeType, field_node: FieldNode) -> GraphQLField:
    try:
        from graphql.execution.execute import get_field_def  # noqa: PLC0415

        return get_field_def(schema, parent_type, field_node)

    # graphql-core >= 3.3.0
    except ImportError:
        return schema.get_field(parent_type=parent_type, field_name=field_node.name.value)


def get_operation_definition(document: DocumentNode, operation_name: str | None) -> OperationDefinitionNode:
    operation_definitions: list[OperationDefinitionNode] = [
        definition_node
        for definition_node in document.definitions
        if isinstance(definition_node, OperationDefinitionNode)
    ]

    if len(operation_definitions) == 0:
        raise GraphQLRequestNoOperationError

    if len(operation_definitions) == 1:
        return operation_definitions[0]

    if operation_name is None:
        raise GraphQLRequestMultipleOperationsNoOperationNameError

    for definition in operation_definitions:
        if definition.name is not None and definition.name.value == operation_name:
            return definition

    raise GraphQLRequestOperationNotFoundError(operation_name=operation_name)


def get_operation_type(document: DocumentNode, operation_name: str | None) -> OperationType:
    operation_definition = get_operation_definition(document, operation_name)
    return operation_definition.operation


def get_error_execution_result(error: GraphQLError | GraphQLErrorGroup | list[GraphQLError]) -> ExecutionResult:
    if isinstance(error, list):
        errors = graphql_errors_hook(error)
        return ExecutionResult(data=None, errors=errors)

    if isinstance(error, GraphQLErrorGroup):
        errors = graphql_errors_hook(list(error.flatten()))
        return ExecutionResult(data=None, errors=errors)

    errors = graphql_errors_hook([error])
    return ExecutionResult(data=None, errors=errors)


# Predicates


def is_connection(field_type: GraphQLOutputType) -> TypeGuard[GraphQLObjectType]:
    return (
        isinstance(field_type, GraphQLObjectType)
        and field_type.name.endswith("Connection")
        and "pageInfo" in field_type.fields
        and "edges" in field_type.fields
    )


def is_edge(field_type: GraphQLOutputType) -> TypeGuard[GraphQLObjectType]:
    return (
        isinstance(field_type, GraphQLObjectType)
        and field_type.name.endswith("Edge")
        and "cursor" in field_type.fields
        and "node" in field_type.fields
    )


def is_node_interface(field_type: GraphQLOutputType) -> TypeGuard[GraphQLInterfaceType]:
    return (
        isinstance(field_type, GraphQLInterfaceType)  # comment here for better formatting
        and field_type.name == "Node"
        and "id" in field_type.fields
    )


def is_page_info(field_type: GraphQLOutputType) -> TypeGuard[GraphQLObjectType]:
    return (
        isinstance(field_type, GraphQLObjectType)
        and field_type.name == "PageInfo"
        and "hasNextPage" in field_type.fields
        and "hasPreviousPage" in field_type.fields
        and "startCursor" in field_type.fields
        and "endCursor" in field_type.fields
    )


def is_typename_metafield(field_node: SelectionNode) -> TypeGuard[FieldNode]:
    if not isinstance(field_node, FieldNode):
        return False
    return field_node.name.value.lower() == "__typename"


def is_relation_id(field: ModelField, field_node: FieldNode) -> TypeGuard[Field]:
    return isinstance(field, ForeignKey) and field.get_attname() == to_snake_case(field_node.name.value)


def should_skip_node(node: NodeWithDirective, variable_values: dict[str, Any]) -> bool:
    skip_args = get_directive_values(GraphQLSkipDirective, node, variable_values)
    if skip_args is not None and skip_args["if"] is True:
        return True

    include_args = get_directive_values(GraphQLIncludeDirective, node, variable_values)
    return include_args is not None and include_args["if"] is False


def is_non_null_default_value(default_value: Any) -> bool:
    return not isinstance(default_value, Hashable) or default_value not in {Undefined, None}


def is_atomic_mutation(operation: OperationDefinitionNode) -> bool:
    """Check if the current operation is an atomic mutation."""
    if operation.operation != OperationType.MUTATION:
        return False

    # Note: String "atomic" is from `GraphQLAtomicDirective.name` but don't import that here just for that
    return "atomic" in {directive.name.value for directive in operation.directives}


# Misc.


async def pre_evaluate_request_user(info: GQLInfo) -> None:
    """
    Fetches the request user from the context and caches it to the request.
    This is a workaround when current user is required in an async event loop,
    but the function itself is not async.
    """
    # '_current_user' would be set by 'django.contrib.auth.middleware.get_user' when calling 'request.user'
    info.context._cached_user = await info.context.auser()  # type: ignore[attr-defined]  # noqa: SLF001


@contextmanager
def graphql_error_path(info: GQLInfo, *, key: str | int | None = None) -> Generator[GQLInfo, None, None]:
    """Context manager that sets the path of all GraphQL errors raised during its context."""
    if key is not None:
        info = GraphQLResolveInfo(  # type: ignore[assignment]
            field_name=info.field_name,
            field_nodes=info.field_nodes,
            return_type=info.return_type,
            parent_type=info.parent_type,
            path=info.path.add_key(key),
            schema=info.schema,
            fragments=info.fragments,
            root_value=info.root_value,
            operation=info.operation,
            variable_values=info.variable_values,
            context=info.context,
            is_awaitable=info.is_awaitable,
        )

    try:
        yield info

    except GraphQLError as error:
        if error.path is None:
            error.path = info.path.as_list()
        raise

    except GraphQLErrorGroup as error_group:
        for err in error_group.flatten():
            if err.path is None:
                err.path = info.path.as_list()
        raise


def check_directives(directives: Iterable[Directive] | None, *, location: DirectiveLocation) -> None:
    """Check that given directives are allowed in the given location."""
    if directives is None:
        return

    for directive in directives:
        if location not in directive.__locations__:
            raise DirectiveLocationError(directive=directive, location=location)


def graphql_errors_hook(errors: list[GraphQLError]) -> list[GraphQLError]:
    """Handle GraphQL errors before adding them to the response."""
    if not errors:
        return errors

    for error in errors:
        extensions: dict[str, Any] = error.extensions

        if error.original_error is None or isinstance(error.original_error, GraphQLError):
            extensions.setdefault("status_code", HTTPStatus.BAD_REQUEST)
        else:
            extensions.setdefault("status_code", HTTPStatus.INTERNAL_SERVER_ERROR)

        if error.__traceback__ is not None:
            log_traceback(error.__traceback__)

            if undine_settings.INCLUDE_ERROR_TRACEBACK:
                extensions["traceback"] = get_traceback(error.__traceback__)

    # Sort the error list in order to make it deterministic
    errors.sort(key=lambda err: (err.locations or [], err.path or [], err.message))
    return errors


def located_validation_error(
    error: ValidationError,
    nodes: Collection[Node],
    path: list[str | int],
) -> GraphQLErrorGroup:
    """Transform a Django ValidationError into a GraphQL errors for each message in the error."""
    code = getattr(error, "code", "").upper()
    error_messages = get_validation_error_messages(error)

    errors: list[GraphQLError] = []
    for field, messages in error_messages.items():
        for message in messages:
            error_path = path.copy()
            if field:
                error_path += field.split(".")

            extensions: dict[str, Any] = {"status_code": HTTPStatus.BAD_REQUEST}
            if code:
                extensions["error_code"] = code

            graphql_error = GraphQLError(message=message, nodes=nodes, path=error_path, extensions=extensions)
            errors.append(graphql_error)

    return GraphQLErrorGroup(errors=errors)



from __future__ import annotations

import re

from django.apps import apps
from django.db.models import CheckConstraint, UniqueConstraint

__all__ = [
    "get_constraint_message",
]


CONSTRAINT_PATTERNS: tuple[re.Pattern, ...] = (
    # Postgres
    re.compile(r'^new row for relation "(?P<relation>\w+)" violates check constraint "(?P<constraint>\w+)"'),
    re.compile(r'^duplicate key value violates unique constraint "(?P<constraint>\w+)"'),
    # SQLite
    re.compile(r"^CHECK constraint failed: (?P<constraint>\w+)$"),
    re.compile(r"^UNIQUE constraint failed: (?P<fields>[\w., ]+)$"),
)


def get_constraint_message(message: str) -> str:
    """Try to get the error message for a constraint violation from the model meta constraints."""
    if (match := CONSTRAINT_PATTERNS[0].match(message)) is not None:
        relation: str = match.group("relation")
        constraint: str = match.group("constraint")
        return postgres_check_constraint_message(relation, constraint, message)

    if (match := CONSTRAINT_PATTERNS[1].match(message)) is not None:
        constraint = match.group("constraint")
        return postgres_unique_constraint_message(constraint, message)

    if (match := CONSTRAINT_PATTERNS[2].match(message)) is not None:
        constraint = match.group("constraint")
        return sqlite_check_constraint_message(constraint, message)

    if (match := CONSTRAINT_PATTERNS[3].match(message)) is not None:
        fields: list[str] = match.group("fields").split(",")
        relation = fields[0].split(".")[0]
        fields = [field.strip().split(".")[1] for field in fields]
        return sqlite_unique_constraint_message(relation, fields, message)

    return message


def postgres_check_constraint_message(relation: str, constraint: str, default_message: str) -> str:
    for model in apps.get_models():
        if model._meta.db_table != relation:
            continue
        for constr in model._meta.constraints:
            if not isinstance(constr, CheckConstraint):
                continue  # pragma: no cover
            if constr.name == constraint:
                return str(constr.violation_error_message)
    return default_message


def postgres_unique_constraint_message(constraint: str, default_message: str) -> str:
    for model in apps.get_models():
        for constr in model._meta.constraints:
            if not isinstance(constr, UniqueConstraint):
                continue  # pragma: no cover
            if constr.name == constraint:
                return str(constr.violation_error_message)
    return default_message


def sqlite_check_constraint_message(constraint: str, default_message: str) -> str:
    for model in apps.get_models():
        for constr in model._meta.constraints:
            if not isinstance(constr, CheckConstraint):
                continue  # pragma: no cover
            if constr.name == constraint:
                return str(constr.violation_error_message)
    return default_message


def sqlite_unique_constraint_message(relation: str, fields: list[str], default_message: str) -> str:
    for model in apps.get_models():
        if model._meta.db_table != relation:
            continue
        for constr in model._meta.constraints:
            if not isinstance(constr, UniqueConstraint):
                continue  # pragma: no cover
            if set(constr.fields) == set(fields):
                return str(constr.violation_error_message)
    return default_message



from __future__ import annotations

import inspect
from collections.abc import AsyncIterable
from types import FunctionType
from typing import Any

from graphql import GraphQLFieldResolver

from undine import Entrypoint, InterfaceType, MutationType, QueryType, UnionType
from undine.converters import convert_to_entrypoint_resolver
from undine.exceptions import InvalidEntrypointMutationTypeError
from undine.pagination import OffsetPagination
from undine.parsers import parse_return_annotation
from undine.relay import Connection, Node
from undine.resolvers import (
    BulkCreateResolver,
    BulkDeleteResolver,
    BulkUpdateResolver,
    ConnectionResolver,
    CreateResolver,
    DeleteResolver,
    EntrypointFunctionResolver,
    InterfaceTypeResolver,
    NodeResolver,
    QueryTypeManyResolver,
    QueryTypeSingleResolver,
    SubscriptionValueResolver,
    UnionTypeResolver,
    UpdateResolver,
)
from undine.resolvers.query import InterfaceTypeConnectionResolver, UnionTypeConnectionResolver
from undine.subscriptions import SignalSubscription
from undine.typing import MutationKind
from undine.utils.reflection import get_origin_or_noop, is_subclass


@convert_to_entrypoint_resolver.register
def _(ref: FunctionType, **kwargs: Any) -> GraphQLFieldResolver:
    caller: Entrypoint = kwargs["caller"]
    if inspect.isasyncgenfunction(ref):
        return SubscriptionValueResolver()

    if inspect.iscoroutinefunction(ref):
        ann = parse_return_annotation(ref)
        origin = get_origin_or_noop(ann)
        if is_subclass(origin, AsyncIterable):
            return SubscriptionValueResolver()

    return EntrypointFunctionResolver(func=ref, entrypoint=caller)


@convert_to_entrypoint_resolver.register
def _(ref: type[QueryType], **kwargs: Any) -> GraphQLFieldResolver:
    caller: Entrypoint = kwargs["caller"]
    if caller.many:
        return QueryTypeManyResolver(query_type=ref, entrypoint=caller)
    return QueryTypeSingleResolver(query_type=ref, entrypoint=caller)


@convert_to_entrypoint_resolver.register
def _(ref: type[MutationType], **kwargs: Any) -> GraphQLFieldResolver:  # noqa: PLR0911
    caller: Entrypoint = kwargs["caller"]

    match ref.__kind__:
        case MutationKind.create:
            if caller.many:
                return BulkCreateResolver(mutation_type=ref, entrypoint=caller)
            return CreateResolver(mutation_type=ref, entrypoint=caller)

        case MutationKind.update:
            if caller.many:
                return BulkUpdateResolver(mutation_type=ref, entrypoint=caller)
            return UpdateResolver(mutation_type=ref, entrypoint=caller)

        case MutationKind.delete:
            if caller.many:
                return BulkDeleteResolver(mutation_type=ref, entrypoint=caller)
            return DeleteResolver(mutation_type=ref, entrypoint=caller)

        case MutationKind.custom:
            if "pk" in ref.__input_map__:
                if caller.many:
                    return BulkUpdateResolver(mutation_type=ref, entrypoint=caller)
                return UpdateResolver(mutation_type=ref, entrypoint=caller)

            if caller.many:
                return BulkCreateResolver(mutation_type=ref, entrypoint=caller)
            return CreateResolver(mutation_type=ref, entrypoint=caller)

        case _:
            raise InvalidEntrypointMutationTypeError(ref=ref, kind=ref.__kind__)


@convert_to_entrypoint_resolver.register
def _(ref: type[UnionType], **kwargs: Any) -> GraphQLFieldResolver:
    caller: Entrypoint = kwargs["caller"]
    return UnionTypeResolver(union_type=ref, entrypoint=caller)


@convert_to_entrypoint_resolver.register
def _(ref: Connection, **kwargs: Any) -> GraphQLFieldResolver:
    caller: Entrypoint = kwargs["caller"]

    if ref.union_type is not None:
        return UnionTypeConnectionResolver(connection=ref, entrypoint=caller)

    if ref.interface_type is not None:
        return InterfaceTypeConnectionResolver(connection=ref, entrypoint=caller)

    return ConnectionResolver(connection=ref, entrypoint=caller)


@convert_to_entrypoint_resolver.register
def _(ref: OffsetPagination, **kwargs: Any) -> GraphQLFieldResolver:
    caller: Entrypoint = kwargs["caller"]

    if ref.union_type is not None:
        return UnionTypeResolver(union_type=ref.union_type, entrypoint=caller)

    if ref.interface_type is not None:
        return InterfaceTypeResolver(interface=ref.interface_type, entrypoint=caller)

    return QueryTypeManyResolver(query_type=ref.query_type, entrypoint=caller)


@convert_to_entrypoint_resolver.register
def _(_: type[Node], **kwargs: Any) -> GraphQLFieldResolver:
    caller: Entrypoint = kwargs["caller"]
    return NodeResolver(entrypoint=caller)


@convert_to_entrypoint_resolver.register
def _(ref: type[InterfaceType], **kwargs: Any) -> GraphQLFieldResolver:
    caller: Entrypoint = kwargs["caller"]
    return InterfaceTypeResolver(interface=ref, entrypoint=caller)


@convert_to_entrypoint_resolver.register
def _(_: SignalSubscription, **kwargs: Any) -> GraphQLFieldResolver:
    return SubscriptionValueResolver()



from __future__ import annotations

from collections.abc import AsyncIterator
from typing import TYPE_CHECKING

from graphql import GraphQLError

from undine.dataclasses import CompletedEventDC, NextEventDC
from undine.exceptions import GraphQLErrorGroup, GraphQLUnexpectedError
from undine.execution import execute_graphql_with_subscription
from undine.typing import DjangoRequestProtocol
from undine.utils.graphql.utils import get_error_execution_result

if TYPE_CHECKING:
    from graphql import ExecutionResult

    from undine.dataclasses import GraphQLHttpParams
    from undine.typing import DjangoRequestProtocol

__all__ = [
    "execute_graphql_sse_dc",
    "result_to_sse_dc",
]


# Distinct connections mode


async def execute_graphql_sse_dc(
    params: GraphQLHttpParams,
    request: DjangoRequestProtocol,
) -> AsyncIterator[NextEventDC | CompletedEventDC]:
    """Execute a GraphQL operation received through server-sent events in distinct connections mode."""
    stream = await execute_graphql_with_subscription(params, request)

    if not isinstance(stream, AsyncIterator):
        yield NextEventDC(data=stream.formatted)
        yield CompletedEventDC()
        return

    try:
        async for data in stream:
            yield NextEventDC(data=data.formatted)

    except GraphQLError as error:
        result = get_error_execution_result(error)
        yield NextEventDC(data=result.formatted)

    except GraphQLErrorGroup as error:
        result = get_error_execution_result(error)
        yield NextEventDC(data=result.formatted)

    except Exception as error:  # noqa: BLE001
        result = get_error_execution_result(GraphQLUnexpectedError(message=str(error)))
        yield NextEventDC(data=result.formatted)

    yield CompletedEventDC()


async def result_to_sse_dc(result: ExecutionResult) -> AsyncIterator[NextEventDC | CompletedEventDC]:  # noqa: RUF029
    """Get iterator for a single result received through server-sent events in distinct connections mode."""
    yield NextEventDC(data=result.formatted)
    yield CompletedEventDC()


# TODO: Single connections mode



"""
File upload handling utilities.

Compliant with the GraphQL multipart request specification:
https://github.com/jaydenseric/graphql-multipart-request-spec
"""

from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Any

from django.core.files import File

from undine.exceptions import GraphQLFileNotFoundError, GraphQLFilePlacingError

if TYPE_CHECKING:
    from django.core.files.uploadedfile import UploadedFile

__all__ = [
    "extract_files",
    "place_files",
]


def extract_files(variables: dict[str, Any]) -> dict[File, list[str]]:
    """
    Extract Django File objects paths in the given variables. Replace the file objects with None values.

    >>> file_1 = File(file=..., name="test_file_1.png")
    >>> file_2 = File(file=..., name="test_file_2.png")
    >>>
    >>> variables = {
    ...     "image": file_1,
    ...     "foo": [file_1, file_2],
    ...     "bar": {"one": file_2, "two": file_1},
    ... }
    >>> files = extract_files(variables)
    >>> files
    {
        <File: test_file_1.png>: ["variables.image", "variables.foo.0", "variables.bar.two"],
        <File: test_file_2.png>: ["variables.foo.1", "variables.bar.one"],
    }
    >>> variables
    {
        "image": None,
        "foo": [None, None],
        "bar": {"one": None, "two": None}
    }
    """
    files: dict[File, list[str]] = defaultdict(list)
    _extract_files_recursively(variables, variables, ["variables"], files)
    return files


def _extract_files_recursively(
    value: Any,
    data: dict[str, Any] | list[Any],
    path: list[str | int],
    files: dict[File, list[str]],
) -> None:
    if isinstance(value, File):
        key = path[-1]
        data[key] = None
        files[value].append(".".join(str(part) for part in path))

    elif isinstance(value, dict):
        for name, item in value.items():
            path.append(name)
            _extract_files_recursively(item, value, path, files)
            path.pop()

    elif isinstance(value, list):
        for index, item in enumerate(value):
            path.append(index)
            _extract_files_recursively(item, value, path, files)
            path.pop()


def place_files(operations: dict[str, Any], files_map: dict[str, list[str]], files: dict[str, UploadedFile]) -> None:
    """
    Place files in the given operations using the files map.

    >>> file_1 = UploadedFile(file=..., name="file_1.png", content_type="image/png")
    >>> file_2 = UploadedFile(file=..., name="file_2.png", content_type="image/png")
    >>>
    >>> operations = {
    ...     "image": None,
    ...     "foo": [None, None],
    ...     "bar": {"one": None, "two": None},
    ... }
    >>> files_map = {
    ...     "0": ["image", "foo.0", "bar.two"]
    ...     "1": ["foo.1", "bar.one"],
    ... }
    ...
    >>> files = {
    ...     "0": file_1,
    ...     "1": file_2,
    ... }
    >>> place_files(operations, files_map, files)
    >>> operations
    {
        "image": file_1,
        "foo": [file_1, file_2],
        "bar": {"one": file_2, "two": file_1},
    }
    """
    for key, values in files_map.items():
        file = files.get(key)
        if file is None:
            raise GraphQLFileNotFoundError(key=key)

        for value in values:
            _place_file(file, value, operations)


def _place_file(file: UploadedFile, value: str, operations: dict[str, Any] | list[Any]) -> None:
    """Handle placing a single file to a single path in the `operations` object."""
    path: list[str] = value.split(".")

    data: Any = operations

    while path:
        key: str | int = path.pop(0)

        try:
            match data:
                case dict():
                    nested_data = data[key]
                case list():
                    key = int(key)
                    nested_data = data[key]
                case _:
                    raise GraphQLFilePlacingError(value=value)

        except (KeyError, IndexError, TypeError) as error:
            raise GraphQLFilePlacingError(value=value) from error

        if path:
            data = nested_data
            continue

        if nested_data is not None:
            raise GraphQLFilePlacingError(value=value)

        data[key] = file



from __future__ import annotations

from .definitions import (
    convert_lookup_to_graphql_type,
    convert_model_field_to_python_type,
    convert_to_default_value,
    convert_to_description,
    convert_to_entrypoint_ref,
    convert_to_entrypoint_resolver,
    convert_to_entrypoint_subscription,
    convert_to_field_complexity,
    convert_to_field_ref,
    convert_to_field_resolver,
    convert_to_filter_lookups,
    convert_to_filter_ref,
    convert_to_filter_resolver,
    convert_to_graphql_argument_map,
    convert_to_graphql_type,
    convert_to_input_ref,
    convert_to_order_ref,
    extend_expression,
    is_field_nullable,
    is_input_hidden,
    is_input_only,
    is_input_required,
    is_many,
)

__all__ = [
    "convert_lookup_to_graphql_type",
    "convert_model_field_to_python_type",
    "convert_to_default_value",
    "convert_to_description",
    "convert_to_entrypoint_ref",
    "convert_to_entrypoint_resolver",
    "convert_to_entrypoint_subscription",
    "convert_to_field_complexity",
    "convert_to_field_ref",
    "convert_to_field_resolver",
    "convert_to_filter_lookups",
    "convert_to_filter_ref",
    "convert_to_filter_resolver",
    "convert_to_graphql_argument_map",
    "convert_to_graphql_type",
    "convert_to_input_ref",
    "convert_to_order_ref",
    "extend_expression",
    "is_field_nullable",
    "is_input_hidden",
    "is_input_only",
    "is_input_required",
    "is_many",
]



from __future__ import annotations

import dataclasses
from collections import defaultdict
from functools import partial
from typing import TYPE_CHECKING, Any, Self, overload

from django.db import transaction  # noqa: ICN003
from django.db.models import Q

from undine.exceptions import (
    GraphQLInvalidInputDataError,
    GraphQLMutationTreeModelMismatchError,
    GraphQLRelationMultipleInstancesError,
    GraphQLRelationNotNullableError,
)
from undine.parsers import parse_model_relation_info
from undine.settings import undine_settings
from undine.typing import RelatedAction, RelationType
from undine.utils.model_utils import (
    generic_relations_for_generic_foreign_key,
    get_bulk_create_kwargs,
    get_default_manager,
    get_instance_or_raise,
    get_instances_or_raise,
    set_forward_ids,
    use_delete_signals,
    use_m2m_add_signals,
    use_m2m_remove_signals,
    use_save_signals,
)
from undine.utils.text import to_camel_case

if TYPE_CHECKING:
    from collections.abc import Callable

    from django.contrib.contenttypes.fields import GenericForeignKey
    from django.db.models import Model
    from django.db.models.fields.related_descriptors import ForeignKeyDeferredAttribute, ManyToManyDescriptor

    from undine.dataclasses import RelInfo
    from undine.typing import TModel

__all__ = [
    "mutate",
]


@overload
def mutate(
    *,
    model: type[TModel],
    data: dict[str, Any],
    related_action: RelatedAction = ...,
) -> TModel: ...


@overload
def mutate(
    *,
    model: type[TModel],
    data: list[dict[str, Any]],
    related_action: RelatedAction = ...,
) -> list[TModel]: ...


@transaction.atomic
def mutate(
    *,
    model: type[TModel],
    data: dict[str, Any] | list[dict[str, Any]],
    related_action: RelatedAction = RelatedAction.null,
) -> TModel | list[TModel]:
    """
    Mutates a instance(s) of the given model using the given input data.

    New instance can created like this:

    >>> instance = mutate(model=Task, data={"name": "New task"})
    >>> instance.name
    'New task'

    Existing instances are used if input data contains the "pk" field.

    >>> instance = mutate(model=Task, data={"pk": 1, "name": "Updated task"})
    >>> instance.name
    'Updated task'

    Related objects can be created, updated or linked at the same time.
    Acceptable values are: a dict containing the input data, the primary key of an existing instance,
    the actual model instance, or None (if null relation is allowed). For many-relations,
    lists of these values are required instead.

    >>> instance = mutate(model=Task, data={"name": "New task", "project": {"name": "New project"}})
    >>> instance.project.name
    'New project'

    If the input data is a list, a list of instances is created.

    >>> instances = mutate(model=Task, data=[{"name": "New task"}, {"name": "New task 2"}])
    >>> len(instances)
    2
    >>> instances[0].name
    'New task'
    >>> instances[1].name
    'New task 2'

    :param model: The model to mutate.
    :param data: The input data to use for the mutation.
    :param related_action: The action to take for existing related objects that are not included in the input.
                           Specifically used for reverse one-to-one and reverse one-to-many relations.
    """
    if isinstance(data, list):
        start_node = MutationNode(model=model, related_action=related_action)
        start_node.handle_many(data)
        return start_node.mutate()  # type: ignore[return-value]

    start_node = MutationNode(model=model, related_action=related_action)
    start_node.handle_one(data)
    return start_node.mutate()[0]  # type: ignore[return-value]


@dataclasses.dataclass(slots=True, kw_only=True)
class MutationNode:
    """
    A node in a tree of mutations.
    Contains instances of a model that should be updated or created.
    May link to other `MutationNodes` to create a tree of mutations
    that can be mutated efficiently.
    """

    model: type[Model]
    """The model that this node is for."""

    related_action: RelatedAction = RelatedAction.null
    """What happens to related objects that are not included in the input?"""

    mutation_func: Callable[[], list[Model]] = dataclasses.field(init=False)
    """The function that should be run to mutate the instances."""

    instances: list[Model] = dataclasses.field(default_factory=list)
    """The instances that should be updated or created."""

    field_names: set[str] = dataclasses.field(default_factory=set)
    """Fields that are present in the input data."""

    before: dict[str, MutationNode] = dataclasses.field(default_factory=dict)
    """Mutations that should be run before this one."""

    after: dict[str, MutationNode] = dataclasses.field(default_factory=dict)
    """Mutations that should be run after this one."""

    def __post_init__(self) -> None:
        self.mutation_func = self.mutate_bulk

    # Mutation handling

    def mutate(self, *, previous_node: MutationNode | None = None) -> list[Model]:
        """Run the mutations starting from this node."""
        for before_node in self.before.values():
            if before_node != previous_node:
                before_node.mutate(previous_node=self)

        instances = self.mutation_func()

        for after_node in self.after.values():
            if after_node != previous_node:
                after_node.mutate(previous_node=self)

        return instances

    def mutate_bulk(self) -> list[Model]:
        """Mutate model instances using the `queryset.bulk_create` method."""
        kwargs = get_bulk_create_kwargs(self.model, *self.field_names)

        instances: list[Model] = []
        for instance in self.instances:
            # Only save new instances if no fields to update
            if kwargs.update_fields or instance.pk is None:
                set_forward_ids(instance)
                if undine_settings.MUTATION_FULL_CLEAN:
                    instance.full_clean()
                instances.append(instance)

        with use_save_signals(self.model, instances, kwargs.update_fields):
            get_default_manager(self.model).bulk_create(objs=instances, **kwargs)

        return self.instances

    def mutate_delete(self) -> list[Model]:
        """Delete model instance s using the `queryset.delete` method."""
        pks = [instance.pk for instance in self.instances]
        with use_delete_signals(self.model, self.instances):
            get_default_manager(self.model).filter(pk__in=pks).delete()
        return []

    def mutate_through(self, *, source_name: str, target_name: str, reverse: bool, symmetrical: bool) -> list[Model]:
        through_map: defaultdict[Model, dict[Model, Model]] = defaultdict(dict)

        # Add new instances
        for through_instance in self.instances:
            set_forward_ids(through_instance)
            if undine_settings.MUTATION_FULL_CLEAN:
                through_instance.full_clean()

            source = getattr(through_instance, source_name)
            target = getattr(through_instance, target_name)
            through_map[source][target] = through_instance

        # Order matters here, since '_upsert_through()' will backfill symmetrical instances to the 'through_map'
        # but '_remove_through()' should not remove any rows in the backwards direction that are
        # not updated during this mutation.
        self._remove_through(
            through_map,
            source_name=source_name,
            target_name=target_name,
            reverse=reverse,
            symmetrical=symmetrical,
        )
        self._upsert_through(
            through_map,
            source_name=source_name,
            target_name=target_name,
            reverse=reverse,
            symmetrical=symmetrical,
        )
        return self.instances

    # Data handling

    def handle_one(self, data: dict[str, Any]) -> Model:
        """Handle data for a single object."""
        pk = data.get("pk")
        # We need to fetch the existing instance so that non-updated fields are present during bulk-create.
        instance = self.model() if pk is None else get_instance_or_raise(model=self.model, pk=pk)

        self.instances.append(instance)
        self.handle_data(data, instance)
        return instance

    def handle_many(self, data: list[dict[str, Any]]) -> list[Model]:
        """Handle data for many objects."""
        instances: list[Model] = []

        instance_map: dict[Any, Model] = {}
        pks = [item["pk"] for item in data if "pk" in item]
        if pks:
            # We need to fetch existing instances so that non-updated fields are present during bulk-create.
            instance_map = {inst.pk: inst for inst in get_instances_or_raise(model=self.model, pks=pks)}

        for item in data:
            pk = item.get("pk")
            instance = instance_map.get(pk) or self.model()

            self.instances.append(instance)
            self.handle_data(item, instance)
            instances.append(instance)

        return instances

    def handle_data(self, data: dict[str, Any], instance: Model) -> Self:
        relation_info = parse_model_relation_info(model=self.model)

        for field_name, field_data in data.items():
            rel_info = relation_info.get(field_name)
            if rel_info is None:
                self.field_names.add(field_name)
                setattr(instance, field_name, field_data)
                continue

            node = MutationNode(model=rel_info.related_model, related_action=self.related_action)  # type: ignore[arg-type]
            self.handle_relation(field_data, rel_info, instance, node)

        return self

    # Relation handing

    def handle_relation(self, data: Any, rel_info: RelInfo, instance: Model, node: MutationNode) -> None:
        match rel_info.relation_type:
            case RelationType.FORWARD_ONE_TO_ONE:
                self._handle_forward_o2o(data, rel_info, instance, node)

            case RelationType.FORWARD_MANY_TO_ONE:
                self._handle_m2o(data, rel_info, instance, node)

            case RelationType.FORWARD_MANY_TO_MANY:
                self._handle_m2m(data, rel_info, instance, node)

            case RelationType.REVERSE_ONE_TO_ONE:
                self._handle_reverse_o2o(data, rel_info, instance, node)

            case RelationType.REVERSE_ONE_TO_MANY:
                self._handle_o2m(data, rel_info, instance, node)

            case RelationType.REVERSE_MANY_TO_MANY:
                self._handle_m2m(data, rel_info, instance, node)

            case RelationType.GENERIC_ONE_TO_MANY:
                self._handle_o2m(data, rel_info, instance, node)

            case RelationType.GENERIC_MANY_TO_ONE:
                self._handle_generic_fk(data, rel_info, instance, node)

            case _:  # pragma: no cover
                msg = f"Unhandled relation type: {rel_info.relation_type}"
                raise NotImplementedError(msg)

    def _handle_forward_o2o(self, data: Any, rel_info: RelInfo, instance: Model, node: MutationNode) -> None:
        """Handle forward one-to-one relations for the given instance."""
        match data:
            case dict():
                rel = node.handle_one(data=data)

                setattr(instance, rel_info.field_name, rel)
                self.field_names.add(rel_info.field_name)  # type: ignore[arg-type]

            case rel_info.related_model_pk_type():
                rel = get_instance_or_raise(model=node.model, pk=data)

                node.instances.append(rel)

                setattr(instance, rel_info.field_name, rel)
                self.field_names.add(rel_info.field_name)  # type: ignore[arg-type]

            case rel_info.related_model():
                node.instances.append(data)

                setattr(instance, rel_info.field_name, data)
                self.field_names.add(rel_info.field_name)  # type: ignore[arg-type]

            case None:
                if not rel_info.related_nullable:
                    raise GraphQLRelationNotNullableError(field_name=rel_info.field_name, model=self.model)

                setattr(instance, rel_info.field_name, None)
                self.field_names.add(rel_info.field_name)  # type: ignore[arg-type]

            case _:
                raise GraphQLInvalidInputDataError(field_name=rel_info.field_name, data=data)

        self.put_before(node, field_name=rel_info.field_name, related_name=rel_info.related_name)  # type: ignore[arg-type]

    _handle_m2o = _handle_forward_o2o
    """Handle many-to-one relations for the given instance."""

    def _handle_reverse_o2o(self, data: Any, rel_info: RelInfo, instance: Model, node: MutationNode) -> None:
        """Handle reverse one-to-one relations for the given instance."""
        existing_instance: Model | None = getattr(instance, rel_info.field_name, None)

        match data:
            case dict():
                rel = node.handle_one(data=data)

                setattr(rel, rel_info.related_name, instance)  # type: ignore[arg-type]
                node.field_names.add(rel_info.related_name)  # type: ignore[arg-type]

            case rel_info.related_model_pk_type():
                rel = get_instance_or_raise(model=node.model, pk=data)

                node.instances.append(rel)

                setattr(rel, rel_info.related_name, instance)  # type: ignore[arg-type]
                node.field_names.add(rel_info.related_name)  # type: ignore[arg-type]

            case rel_info.related_model():
                node.instances.append(data)

                setattr(data, rel_info.related_name, instance)  # type: ignore[arg-type]
                node.field_names.add(rel_info.related_name)  # type: ignore[arg-type]

            case None:
                # If creating, no related instance is created.
                # If updating, there might be a related instance.
                # What happens to it depends on the "related action" set by the mutation type.
                pass

            case _:
                raise GraphQLInvalidInputDataError(field_name=rel_info.field_name, data=data)

        self.put_after(node, field_name=rel_info.field_name, related_name=rel_info.related_name)  # type: ignore[arg-type]

        if existing_instance is not None:
            updated_instances: set[Model] = {instance for instance in node.instances if instance.pk is not None}
            non_updated_instances: set[Model] = {existing_instance} - updated_instances

            if non_updated_instances:
                match self.related_action:
                    case RelatedAction.null:
                        self._disconnect_instances(non_updated_instances, rel_info, node)
                    case RelatedAction.delete:
                        self._remove_instances(non_updated_instances, rel_info, node)
                    case RelatedAction.ignore:
                        raise GraphQLRelationMultipleInstancesError(
                            field_name=rel_info.related_name,
                            model=rel_info.related_model,
                        )

    def _handle_o2m(self, data: list[Any], rel_info: RelInfo, instance: Model, node: MutationNode) -> None:  # noqa: C901,PLR0912
        """Handle one-to-many relations for the given instance."""
        existing_instances: set[Model] = set()
        if instance.pk is not None:
            existing_instances = set(getattr(instance, rel_info.field_name).all())

        match data:
            case [dict(), *_]:
                instances = node.handle_many(data=data)

                for rel in instances:
                    setattr(rel, rel_info.related_name, instance)  # type: ignore[arg-type]

                node.field_names.add(rel_info.related_name)  # type: ignore[arg-type]

            case [rel_info.related_model_pk_type(), *_]:
                for rel in get_instances_or_raise(model=node.model, pks=data):
                    node.instances.append(rel)
                    setattr(rel, rel_info.related_name, instance)  # type: ignore[arg-type]

                node.field_names.add(rel_info.related_name)  # type: ignore[arg-type]

            case [rel_info.related_model(), *_]:
                for rel in data:
                    node.instances.append(rel)
                    setattr(rel, rel_info.related_name, instance)  # type: ignore[arg-type]

                node.field_names.add(rel_info.related_name)  # type: ignore[arg-type]

            case [None, *_] | [] | None:
                # If creating, do not add related instances.
                # If updating, all instances will be "non-updated" instances.
                # What happens to them depends on the "related action" set by the mutation type.
                pass

            case _:
                raise GraphQLInvalidInputDataError(field_name=rel_info.field_name, data=data)

        self.put_after(node, field_name=rel_info.field_name, related_name=rel_info.related_name)  # type: ignore[arg-type]

        if existing_instances:
            updated_instances: set[Model] = {instance for instance in node.instances if instance.pk is not None}
            non_updated_instances: set[Model] = existing_instances - updated_instances

            if non_updated_instances:
                match self.related_action:
                    case RelatedAction.null:
                        self._disconnect_instances(non_updated_instances, rel_info, node)
                    case RelatedAction.delete:
                        self._remove_instances(non_updated_instances, rel_info, node)
                    case RelatedAction.ignore:
                        pass

    def _handle_m2m(self, data: list[Any], rel_info: RelInfo, instance: Model, node: MutationNode) -> None:
        """Handle many-to-many relations for the given instance."""
        match data:
            case [dict(), *_]:
                node.handle_many(data=data)

            case [rel_info.related_model_pk_type(), *_]:
                for rel in get_instances_or_raise(model=node.model, pks=data):
                    node.instances.append(rel)

            case [rel_info.related_model(), *_]:
                for rel in data:
                    node.instances.append(rel)

            case [None, *_] | [] | None:
                node.instances = []

            case _:
                raise GraphQLInvalidInputDataError(field_name=rel_info.field_name, data=data)

        node._handle_through(instance, rel_info)

        self.put_after(node, field_name=rel_info.field_name, related_name=rel_info.related_name)  # type: ignore[arg-type]

    def _handle_through(self, source: Model, rel_info: RelInfo) -> None:
        """Handle the through model instances for a many-to-many relation."""
        m2m: ManyToManyDescriptor = getattr(type(source), rel_info.field_name)
        reverse = rel_info.relation_type.is_reverse
        source_name = m2m.field.m2m_reverse_field_name() if reverse else m2m.field.m2m_field_name()
        target_name = m2m.field.m2m_field_name() if reverse else m2m.field.m2m_reverse_field_name()
        symmetrical = m2m.rel.symmetrical

        node = MutationNode(model=m2m.through, related_action=self.related_action)  # type: ignore[arg-type]
        node.mutation_func = partial(
            node.mutate_through,
            source_name=source_name,
            target_name=target_name,
            reverse=reverse,
            symmetrical=symmetrical,
        )

        for target in self.instances:
            through_instance = node.model()
            setattr(through_instance, source_name, source)
            setattr(through_instance, target_name, target)

            node.instances.append(through_instance)
            node.field_names.add(source_name)
            node.field_names.add(target_name)

        attr: ForeignKeyDeferredAttribute = getattr(m2m.through, source_name)
        hidden_field_name = attr.field.remote_field.get_accessor_name()

        self.put_after(node, field_name=hidden_field_name, related_name=source_name)  # type: ignore[arg-type]

    def _handle_generic_fk(self, data: Any, rel_info: RelInfo, instance: Model, node: MutationNode) -> None:
        """Handle generic foreign key relations for the given instance."""
        if not isinstance(data, dict):
            raise GraphQLInvalidInputDataError(field_name=rel_info.field_name, data=data)

        key: str | None = next(iter(data), None)
        if key is None:
            raise GraphQLInvalidInputDataError(field_name=rel_info.field_name, data=data)

        model_data: dict[str, Any] = data[key]

        field: GenericForeignKey = rel_info.model._meta.get_field(rel_info.field_name)  # type: ignore[assignment]
        relations = generic_relations_for_generic_foreign_key(field)
        related_model_map = {to_camel_case(rel.model.__name__): rel.model for rel in relations}

        model = related_model_map.get(key)
        if model is None:
            msg = f"Model '{key}' doesn't exist or have a generic relation to '{rel_info.model.__name__}'."
            raise GraphQLInvalidInputDataError(msg)

        node.model = model

        match model_data:
            case dict():
                rel = node.handle_one(data=model_data)

                setattr(instance, rel_info.field_name, rel)
                self.field_names.add(rel_info.field_name)  # type: ignore[arg-type]

            case None:
                if not rel_info.related_nullable:
                    raise GraphQLRelationNotNullableError(field_name=rel_info.field_name, model=self.model)

                setattr(instance, rel_info.field_name, None)
                self.field_names.add(rel_info.field_name)  # type: ignore[arg-type]

            case _:
                raise GraphQLInvalidInputDataError(field_name=rel_info.field_name, data=model_data)

        self.put_before(node, field_name=rel_info.field_name, related_name=rel_info.related_name)  # type: ignore[arg-type]

    def _disconnect_instances(self, instances: set[Model], rel_info: RelInfo, node: MutationNode) -> None:
        """
        Disconnect instances for:

        - reverse one-to-one, if the relation is updated to another instance
        - reverse foreign keys relations, if a some of the instances are not updated or picked using pk.

        Do not allow disconnecting if the forward relation is not nullable.
        """
        if not rel_info.related_nullable:
            raise GraphQLRelationNotNullableError(field_name=rel_info.related_name, model=node.model)

        disconnect_node = MutationNode(model=rel_info.related_model, related_action=self.related_action)  # type: ignore[arg-type]

        for instance in instances:
            setattr(instance, rel_info.related_name, None)  # type: ignore[arg-type]
            disconnect_node.instances.append(instance)
            disconnect_node.field_names.add(rel_info.related_name)  # type: ignore[arg-type]

        # For reverse one-to-one relations, existing relation must be disconnected before new relation is added
        # to satisfy on-to-one constraint.
        node.put_before(
            disconnect_node,
            field_name=f"__disconnect_old_{rel_info.field_name}",
            related_name=f"__connect_new_{rel_info.field_name}",
        )

    def _remove_instances(self, instances: set[Model], rel_info: RelInfo, node: MutationNode) -> None:
        """Remove the given related instances when a relation is updated."""
        remove_node = MutationNode(model=rel_info.related_model, related_action=self.related_action)  # type: ignore[arg-type]
        remove_node.instances.extend(instances)
        remove_node.mutation_func = remove_node.mutate_delete

        # For reverse one-to-one relations, existing relation must be removed before new relation is added
        # to satisfy on-to-one constraint.
        node.put_before(
            remove_node,
            field_name=f"__remove_old_{rel_info.field_name}",
            related_name=f"__add_new_{rel_info.field_name}",
        )

    # Link handling

    def merge(self, node: MutationNode, *, previous_node: MutationNode | None = None) -> Self:
        """Merge the given MutationNode into this one."""
        if node.model != self.model:
            raise GraphQLMutationTreeModelMismatchError(model_1=node.model, model_2=self.model)

        self.instances.extend(node.instances)

        for name, before_node in self.before.items():
            if previous_node == before_node:
                continue

            other_before_node = node.before.get(name)
            if other_before_node is not None:
                before_node.merge(other_before_node, previous_node=self)

        for name, after_node in self.after.items():
            if previous_node == after_node:
                continue

            other_after_node = node.after.get(name)
            if other_after_node is not None:
                after_node.merge(other_after_node, previous_node=self)

        return self

    def put_before(self, node: MutationNode, *, field_name: str, related_name: str) -> Self:
        """Put the given MutationNode before the current one."""
        before_node = self.before.get(field_name)
        if before_node is not None:
            before_node.merge(node)
            return self

        self.before[field_name] = node
        node.after[related_name] = self
        return self

    def put_after(self, node: MutationNode, *, field_name: str, related_name: str) -> Self:
        """Put the given MutationNode after the current one."""
        after_node = self.after.get(field_name)
        if after_node is not None:
            after_node.merge(node)
            return self

        self.after[field_name] = node
        node.before[related_name] = self
        return self

    # Through model handling

    def _remove_through(
        self,
        through_map: defaultdict[Model, dict[Model, Model]],
        *,
        source_name: str,
        target_name: str,
        reverse: bool,
        symmetrical: bool,
    ) -> None:
        """Remove through model instances."""
        not_updated = Q()
        for source, target_map in through_map.items():
            not_updated |= Q(**{source_name: source}) & ~Q(**{f"{target_name}__in": list(target_map)})
            if symmetrical:
                not_updated |= Q(**{target_name: source}) & ~Q(**{f"{source_name}__in": list(target_map)})

        pks: set[Any] = set()
        source_to_removed_target_pks: defaultdict[Model, set[Any]] = defaultdict(set)

        qs = get_default_manager(self.model).filter(not_updated).select_related(source_name, target_name)

        for through_instance in qs:
            source = getattr(through_instance, source_name)
            target = getattr(through_instance, target_name)
            source_to_removed_target_pks[source].add(target.pk)
            pks.add(through_instance.pk)

        # If there are no through instances to remove, we can skip the rest
        if not pks:
            return

        with use_m2m_remove_signals(
            model=self.model,
            source_to_removed_target_pks=source_to_removed_target_pks,
            target_name=target_name,
            reverse=reverse,
        ):
            get_default_manager(self.model).filter(pk__in=pks).delete()

    def _upsert_through(
        self,
        through_map: defaultdict[Model, dict[Model, Model]],
        *,
        source_name: str,
        target_name: str,
        reverse: bool,
        symmetrical: bool,
    ) -> None:
        """Add or update through model instances."""
        # If there are no through instances, we can skip the upsert
        if not through_map:
            return

        if symmetrical:
            self._backfill_symmetrical(
                through_map,
                source_name=source_name,
                target_name=target_name,
            )

        self._use_existing_through(
            through_map,
            source_name=source_name,
            target_name=target_name,
            symmetrical=symmetrical,
        )

        instances: list[Model] = []
        source_to_added_target_pks: defaultdict[Model, set[Any]] = defaultdict(set)

        for source, target_map in through_map.items():
            for target, instance in target_map.items():
                instances.append(instance)
                # Send signals only for added instances, not for updated existing ones
                if instance.pk is None:
                    source_to_added_target_pks[source].add(target.pk)

        kwargs = get_bulk_create_kwargs(self.model, *self.field_names)

        with use_m2m_add_signals(
            model=self.model,
            source_to_added_target_pks=source_to_added_target_pks,
            target_name=target_name,
            reverse=reverse,
        ):
            get_default_manager(self.model).bulk_create(objs=instances, **kwargs)

    def _backfill_symmetrical(
        self,
        through_map: defaultdict[Model, dict[Model, Model]],
        *,
        source_name: str,
        target_name: str,
    ) -> None:
        """Add symmetrical instances to the 'through_map'."""
        symmetric_map: defaultdict[Model, dict[Model, Model]] = defaultdict(dict)

        field_names = {field.name for field in self.model._meta.get_fields()}
        field_names.discard(self.model._meta.pk.name)
        field_names.discard(source_name)
        field_names.discard(target_name)

        for source, target_map in through_map.items():
            for target, instance in target_map.items():
                symmetrical_instance = self.model()
                setattr(symmetrical_instance, source_name, target)
                setattr(symmetrical_instance, target_name, source)

                for field_name in field_names:
                    setattr(symmetrical_instance, field_name, getattr(instance, field_name))

                symmetric_map[target][source] = symmetrical_instance

        for target, source_map in symmetric_map.items():
            for source, instance in source_map.items():
                through_map[target][source] = instance

    def _use_existing_through(
        self,
        through_map: defaultdict[Model, dict[Model, Model]],
        *,
        source_name: str,
        target_name: str,
        symmetrical: bool,
    ) -> None:
        """
        Replace new instances in the 'through_map' with existing through model instances
        that have the same source and target.
        """
        existing = Q()
        for source, target_map in through_map.items():
            existing |= Q(**{source_name: source, f"{target_name}__in": list(target_map)})
            if symmetrical:
                existing |= Q(**{target_name: source, f"{source_name}__in": list(target_map)})

        qs = get_default_manager(self.model).filter(existing).select_related(source_name, target_name)

        field_names = {field.name for field in self.model._meta.get_fields()}
        field_names.discard(self.model._meta.pk.name)
        field_names.discard(source_name)
        field_names.discard(target_name)

        for through_instance in qs:
            source = getattr(through_instance, source_name)
            target = getattr(through_instance, target_name)

            new_through_instance = through_map[source][target]
            for field_name in field_names:
                setattr(through_instance, field_name, getattr(new_through_instance, field_name))

            through_map[source][target] = through_instance



from __future__ import annotations

import uuid
from inspect import cleandoc

from ._definition import ScalarType

__all__ = [
    "GraphQLUUID",
    "uuid_scalar",
]


uuid_scalar: ScalarType[uuid.UUID, str] = ScalarType(
    name="UUID",
    description=cleandoc(
        """
        Represents a universally unique identifier string.
        Maps to Python's `uuid.UUID` type.
        """
    ),
    specified_by_url="https://datatracker.ietf.org/doc/html/rfc9562",
)

GraphQLUUID = uuid_scalar.as_graphql_scalar()


@uuid_scalar.parse.register
def _(value: str) -> uuid.UUID:
    return uuid.UUID(hex=value)


@uuid_scalar.parse.register
def _(value: int) -> uuid.UUID:
    return uuid.UUID(int=value)


@uuid_scalar.serialize.register
def _(value: str) -> str:
    return str(uuid.UUID(hex=value))


@uuid_scalar.serialize.register
def _(value: uuid.UUID) -> str:
    return str(value)


@uuid_scalar.serialize.register
def _(value: int) -> str:
    return str(uuid.UUID(int=value))


@uuid_scalar.serialize.register
def _(value: bytes) -> str:
    return str(uuid.UUID(bytes=value))



from __future__ import annotations

from copy import copy
from typing import Any

from django.db.models import Field, Lookup, Transform
from django.db.models.constants import LOOKUP_SEP
from django.db.models.functions import TruncDate, TruncTime
from graphql import GraphQLArgumentMap, GraphQLFieldResolver, GraphQLInputType, GraphQLOutputType

from undine.typing import DjangoExpression, GraphQLFilterResolver, SupportsLookup
from undine.utils.function_dispatcher import FunctionDispatcher
from undine.utils.reflection import has_callable_attribute, is_subclass

__all__ = [
    "convert_lookup_to_graphql_type",
    "convert_model_field_to_python_type",
    "convert_to_bad_lookups",
    "convert_to_default_value",
    "convert_to_description",
    "convert_to_entrypoint_ref",
    "convert_to_entrypoint_resolver",
    "convert_to_field_ref",
    "convert_to_field_resolver",
    "convert_to_filter_lookups",
    "convert_to_filter_ref",
    "convert_to_filter_resolver",
    "convert_to_graphql_argument_map",
    "convert_to_graphql_type",
    "convert_to_input_ref",
    "convert_to_order_ref",
    "extend_expression",
    "is_field_nullable",
    "is_input_hidden",
    "is_input_only",
    "is_input_required",
    "is_many",
]

# Converters are defined here without their implementations to avoid circular imports.
# The implementations are defined in the `impl` subpackage.
# These are then registered in `apps.py` when Django apps are ready.


convert_lookup_to_graphql_type: FunctionDispatcher[GraphQLInputType | GraphQLOutputType] = FunctionDispatcher()
"""
Convert the given Model field lookup to a GraphQL type.

Arguments:

`lookup: str`: The lookup made to the Model field.

`default_type: GraphQLInputType | GraphQLOutputType`: The GraphQL Type for the parent field the lookup is for.
"""

convert_model_field_to_python_type: FunctionDispatcher[type] = FunctionDispatcher()
"""
Convert the given Model field to a python type.

Arguments:

`ref: Any`: The Model field to convert.
"""

convert_to_bad_lookups: FunctionDispatcher[set[str]] = FunctionDispatcher()
"""
The output of `.get_lookups()` may include lookups that don't actually work or are undesirable
for the field. This function can be used to register these lookups per model field so that
FilterSet autogeneration can avoid generating Filters for these lookups.

Arguments:

`field: ModelField`: The Django Model field to look at.
"""

convert_to_default_value: FunctionDispatcher[Any] = FunctionDispatcher()
"""
Convert the given reference to a default value for an Input.
Returns `Undefined` if the reference doesn't have a default value.

Arguments:

`ref: Any`: The reference to convert to a default value.

`caller: Input`: The Input instance that is calling this function.
"""

convert_to_description: FunctionDispatcher[str | None] = FunctionDispatcher()
"""
Convert the given convert.

Arguments:

`ref: Any`: The reference to parse description from.
"""


convert_to_entrypoint_ref: FunctionDispatcher[Any] = FunctionDispatcher()
"""
Convert the given value to a reference that Entrypoint can deal with.

Arguments:

`ref: Any`: The value to convert.

`caller: Entrypoint`: Entrypoint instance that is calling this function.
"""


convert_to_entrypoint_resolver: FunctionDispatcher[GraphQLFieldResolver] = FunctionDispatcher()
"""
Convert the given reference to a GraphQL field resolver for an Entrypoint.

Arguments:

`ref: Any`: The reference to convert.

`caller: Entrypoint`: The Entrypoint instance that is calling this function.
"""

convert_to_entrypoint_subscription: FunctionDispatcher[GraphQLFieldResolver | None] = FunctionDispatcher()
"""
Convert the given reference to a GraphQL subscription resolver for an Entrypoint.

Arguments:

`ref: Any`: The reference to convert.

`caller: Entrypoint`: The Entrypoint instance that is calling this function.
"""

convert_to_field_complexity: FunctionDispatcher[int] = FunctionDispatcher()
"""
Determine the complexity of resolving the given reference.

Arguments:

`ref: Any`: The reference to look at.

`caller: Field`: The Field instance that is calling this function.
"""

convert_to_field_ref: FunctionDispatcher[Any] = FunctionDispatcher()
"""
Convert the given value to a reference that Field can deal with.

Arguments:

`ref: Any`: The value to convert.

`caller: Field`: Field instance that is calling this function.
"""

convert_to_field_resolver: FunctionDispatcher[GraphQLFieldResolver] = FunctionDispatcher()
"""
Convert the given reference to a field resolver function for Field.

Arguments:

`ref: Any`: The reference to convert.

`caller: Field`: The Field instance that is calling this function.
"""

convert_to_filter_ref: FunctionDispatcher[Any] = FunctionDispatcher()
"""
Convert the given value to a reference that Filter can deal with.

Arguments:

`ref: Any`: The value to convert.

`caller: Filter`: The Filter instance that is calling this function.
"""

convert_to_filter_resolver: FunctionDispatcher[GraphQLFilterResolver] = FunctionDispatcher()
"""
Convert the given reference to a filter resolver function for Filter.

Arguments:

`ref: Any`: The reference to convert.

`caller: Filter`: The Filter instance that is calling this function.
"""

convert_to_graphql_argument_map: FunctionDispatcher[GraphQLArgumentMap] = FunctionDispatcher()
"""
Parse a GraphQLArgumentMap from the given Entrypoint or Field reference.

Arguments:

`ref: Any`: The reference to convert.

`many: bool`: Whether the argument map is for a list field.

`entrypoint: bool = False`: (Optional) Whether the argument map is for an entrypoint.
"""

convert_to_graphql_type: FunctionDispatcher[GraphQLInputType | GraphQLOutputType] = FunctionDispatcher()
"""
Convert a given reference to a GraphQL input or output type.

Arguments:

`ref: Any`: The reference to convert.

`model: type[Model]`: The Django Model associated with the reference.

`is_input: bool = False`: (Optional) Whether the converted type should be an input or output type.
"""

convert_to_input_ref: FunctionDispatcher[Any] = FunctionDispatcher()
"""
Convert the given value to a reference that Input can deal with.

Arguments:

`ref: Any`: The value to convert.

`caller: Input`: The Input instance that is calling this function.
"""

convert_to_order_ref: FunctionDispatcher[Any] = FunctionDispatcher()
"""
Convert the given value to a reference that Order can deal with.

Arguments:

`ref: Any`: The value to convert.

`caller: Order`: The Order instance that is calling this function.
"""


extend_expression: FunctionDispatcher[DjangoExpression] = FunctionDispatcher()
"""
Rewrite an expression so that any containing lookups are referenced through the given field.

Arguments:

`expression: DjangoExpression`: The expression to extend.

`field_name: str`: Name of the field to extend the lookup to.
"""

is_field_nullable: FunctionDispatcher[bool] = FunctionDispatcher()
"""
Determine whether the given reference indicates a nullable Field or not.

Arguments:

`ref: Any`: The reference to check.

`caller: Field`: The Field instance that is calling this function.
"""

is_input_hidden: FunctionDispatcher[bool] = FunctionDispatcher()
"""
Determine whether the given reference indicates a hidden Input or not.

Arguments:

`ref: Any`: The reference to check.

`caller: Input`: The Input instance that is calling this function.
"""

is_input_only: FunctionDispatcher[bool] = FunctionDispatcher()
"""
Determine whether the given reference is indicates an input-only Input or not.

Arguments:

`ref: Any`: The reference to check.

`caller: Input`: The Input instance that is calling this function.
"""

is_input_required: FunctionDispatcher[bool] = FunctionDispatcher()
"""
Determine whether the give reference indicates a required Input or not.

Arguments:

`ref: Any`: The reference to check.

`caller: Input`: The Input instance that is calling this function.
"""

is_many: FunctionDispatcher[bool] = FunctionDispatcher()
"""
Determine whether the given reference indicates a list of objects or not.

Arguments:

`ref: Any`: The reference to look at.

`model: type[Model]`: The Django Model associated with the reference.

`name: str`: A name associated with the reference (e.g. field name)
"""


def convert_to_filter_lookups(ref: SupportsLookup, *, path: str = "") -> set[str]:
    """
    Find all lookups reachable from the given reference.
    This includes lookups for any transforms registered for the reference.
    """
    lookups: set[str] = set()
    for name, expr in _get_lookups_and_transforms(ref).items():
        # No need to include 'exact' lookup for nested lookups since the parent lookup does the same thing
        if path and name == "exact":
            continue

        lookup_name = f"{path}{LOOKUP_SEP}{name}" if path else name

        lookups.add(lookup_name)

        if has_callable_attribute(expr, "get_lookups"):
            lookups |= convert_to_filter_lookups(expr, path=lookup_name)  # type: ignore[arg-type]

    return lookups


def _get_lookups_and_transforms(ref: SupportsLookup) -> dict[str, type[Lookup | Transform]]:
    """
    Get all lookups and transforms for the given reference, including its 'output_field', if it has one.
    Exclude regex lookups and 'bad' lookups registered with 'get_bad_lookups'.
    """
    from django.contrib.contenttypes.fields import GenericForeignKey

    # Cannot do lookups on GenericForeignKeys
    if isinstance(ref, GenericForeignKey):
        return {}

    lookups: dict[str, type[Lookup | Transform]] = copy(ref.get_lookups())

    if hasattr(ref, "output_field") and isinstance(ref.output_field, Field):
        for name in _get_lookups_and_transforms(ref.output_field):
            existing_lookup = lookups.get(name)
            if existing_lookup is not None:
                continue

            lookup: type[Lookup | Transform] | None = ref.output_field.get_lookups().get(name, None)
            if lookup is None:  # pragma: no cover
                lookups.pop(name, None)
                continue

            # Don't include transforms for 'TruncDate' and 'TruncTime', since those are only registered
            # on 'DateTimeField', which already has all the lookups of 'DateField' and 'TimeField'.
            # This way the created 'FilterSet' will not be quite so verbose.
            if ref in {TruncDate, TruncTime} and is_subclass(lookup, Transform):
                continue

            lookups[name] = lookup

    # Don't allow regex inputs from users
    lookups.pop("iregex", None)
    lookups.pop("regex", None)

    for bad_lookup in convert_to_bad_lookups(ref):
        lookups.pop(bad_lookup, None)

    return lookups



from __future__ import annotations

import base64

from ._definition import ScalarType

__all__ = [
    "GraphQLBase64",
    "base64_scalar",
]


base64_scalar: ScalarType[str, str] = ScalarType(
    name="Base64",
    description="Represents a base64-encoded string.",
    specified_by_url="https://datatracker.ietf.org/doc/html/rfc4648#section-4",
)

GraphQLBase64 = base64_scalar.as_graphql_scalar()


@base64_scalar.serialize.register
@base64_scalar.parse.register
def _(value: str) -> str:
    # Validate string is base64 encoded
    base64.b64decode(value.encode(encoding="utf-8")).decode(encoding="utf-8")
    return value


@base64_scalar.serialize.register
def _(value: bytes) -> str:
    # Validate string is base64 encoded
    base64.b64decode(value).decode(encoding="utf-8")
    return value.decode(encoding="utf-8")



from __future__ import annotations

import ast
import inspect
import re
from functools import cache
from inspect import cleandoc
from textwrap import dedent

from undine.exceptions import InvalidDocstringParserError
from undine.settings import undine_settings
from undine.typing import DocstringParserProtocol

__all__ = [
    "docstring_parser",
    "parse_class_attribute_docstrings",
]


class PreProcessingDocstringParser:
    """
    A docstring parser that implements another parser with
    pre-processing steps for null-checking and input sanitation operations.
    """

    def __init__(self, parser_impl: type[DocstringParserProtocol]) -> None:
        if not isinstance(parser_impl, DocstringParserProtocol):
            raise InvalidDocstringParserError(cls=parser_impl)

        self.parser_impl = parser_impl

    def parse_body(self, docstring: str | None) -> str | None:
        if docstring is None:
            return None
        docstring = cleandoc(docstring).strip()
        return self.parser_impl.parse_body(docstring)

    def parse_arg_descriptions(self, docstring: str | None) -> dict[str, str]:
        if docstring is None:
            return {}
        docstring = cleandoc(docstring).strip()
        return self.parser_impl.parse_arg_descriptions(docstring)

    def parse_return_description(self, docstring: str | None) -> str | None:
        if docstring is None:
            return None
        docstring = cleandoc(docstring).strip()
        return self.parser_impl.parse_return_description(docstring)

    def parse_raise_descriptions(self, docstring: str | None) -> dict[str, str]:
        if docstring is None:
            return {}
        docstring = cleandoc(docstring).strip()
        return self.parser_impl.parse_raise_descriptions(docstring)

    def parse_deprecations(self, docstring: str | None) -> dict[str, str]:
        if docstring is None:
            return {}
        docstring = cleandoc(docstring).strip()
        return self.parser_impl.parse_deprecations(docstring)


class RSTDocstringParser:
    """A docstring parser for reStructuredText formatted docstrings."""

    TRIM_PATTERN = re.compile(r"\s+")
    STOP_PATTERN = r"(?=:param|:returns?|:raises?|:deprecated?|$)"
    BODY_PATTERN = re.compile(rf"^(?P<text>.+?){STOP_PATTERN}.*", flags=re.DOTALL)
    PARAM_PATTERN = re.compile(rf":param (?P<name>\w+): (?P<text>.+?){STOP_PATTERN}", flags=re.DOTALL)
    RETURN_PATTERN = re.compile(rf":returns?: (?P<text>.+?){STOP_PATTERN}", flags=re.DOTALL)
    RAISE_PATTERN = re.compile(rf":raises? (?P<name>\w+): (?P<text>.+?){STOP_PATTERN}", flags=re.DOTALL)
    DEPR_PATTERN = re.compile(rf":deprecated? (?P<name>\w+): (?P<text>.+?){STOP_PATTERN}", flags=re.DOTALL)

    @classmethod
    def parse_body(cls, docstring: str) -> str:
        body: str = ""
        body_match = cls.BODY_PATTERN.search(docstring)
        if body_match is not None:
            body = body_match.group("text").strip()
        return body

    @classmethod
    def parse_arg_descriptions(cls, docstring: str) -> dict[str, str]:
        found_args = cls.PARAM_PATTERN.findall(docstring)
        return {name: re.sub(cls.TRIM_PATTERN, " ", text).strip() for name, text in found_args}

    @classmethod
    def parse_return_description(cls, docstring: str) -> str:
        return_description: str = ""
        returns_match = cls.RETURN_PATTERN.search(docstring)
        if returns_match is not None:
            return_description = re.sub(cls.TRIM_PATTERN, " ", returns_match.group("text")).strip()
        return return_description

    @classmethod
    def parse_raise_descriptions(cls, docstring: str) -> dict[str, str]:
        found_raises = cls.RAISE_PATTERN.findall(docstring)
        return {key: re.sub(cls.TRIM_PATTERN, " ", value).strip() for key, value in found_raises}

    @classmethod
    def parse_deprecations(cls, docstring: str) -> dict[str, str]:
        found_deprecations = cls.DEPR_PATTERN.findall(docstring)
        return {key: re.sub(cls.TRIM_PATTERN, " ", value).strip() for key, value in found_deprecations}


class PlainDocstringParser:
    """
    A docstring parser for plain docstrings.
    Parses the body as is and leaves everything else alone.
    """

    @classmethod
    def parse_body(cls, docstring: str) -> str:
        return docstring.strip()

    @classmethod
    def parse_arg_descriptions(cls, docstring: str) -> dict[str, str]:
        return {}

    @classmethod
    def parse_return_description(cls, docstring: str) -> str:
        return ""

    @classmethod
    def parse_raise_descriptions(cls, docstring: str) -> dict[str, str]:
        return {}

    @classmethod
    def parse_deprecations(cls, docstring: str) -> dict[str, str]:
        return {}


docstring_parser = PreProcessingDocstringParser(parser_impl=undine_settings.DOCSTRING_PARSER)


@cache
def parse_class_attribute_docstrings(cls: type, /) -> dict[str, str]:
    """
    Parse class attribute docstrings from the given class.

    >>> class Foo:
    ...     bar: int = 1
    ...     '''Description.'''
    ...
    ...     baz: int = 2
    ...     '''Another description.'''
    >>>
    >>> parse_class_attribute_docstrings(Foo)
    {'bar': 'Description.', 'baz': 'Another description.'}
    """
    # Class attribute docstrings are disabled by default to improve performance.
    if not undine_settings.ENABLE_CLASS_ATTRIBUTE_DOCSTRINGS:
        return {}

    descriptions: dict[str, str] = {}

    try:
        module = ast.parse(dedent(inspect.getsource(cls)))
    except OSError:  # Classes created with `type(...)` don't have a source.
        return descriptions

    class_def: ast.ClassDef = module.body[0]

    name: str | None = None
    for expr in class_def.body:
        # Variable docstrings are always directly below the class attribute definition.
        if name is not None:
            if isinstance(expr, ast.Expr) and isinstance(expr.value, ast.Constant):
                descriptions[name] = inspect.cleandoc(expr.value.value)
            name = None
            continue

        if isinstance(expr, ast.AnnAssign):
            name = expr.target.id
        elif isinstance(expr, ast.Assign):
            name = expr.targets[0].id

    return descriptions



from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING, TypeGuard

if TYPE_CHECKING:
    from django.db.models import Field, Model


with suppress(ImportError):
    from modeltranslation.fields import TranslationField
    from modeltranslation.manager import get_translatable_fields_for_model


__all__ = [
    "get_translatable_fields",
    "is_translation_field",
]


IS_MODELTRANSLATION_INSTALLED: bool = "TranslationField" in globals()


def is_translation_field(field: Field) -> TypeGuard[TranslationField]:
    if not IS_MODELTRANSLATION_INSTALLED:
        return False

    return isinstance(field, TranslationField)


def get_translatable_fields(model: type[Model]) -> set[str]:
    """If `django-modeltranslation` is installed, find all translatable fields in the given model."""
    if not IS_MODELTRANSLATION_INSTALLED:
        return set()

    return set(get_translatable_fields_for_model(model) or [])



from __future__ import annotations

import dataclasses
import json
from enum import StrEnum
from typing import TYPE_CHECKING, Any

from django.shortcuts import render

from undine.settings import undine_settings

if TYPE_CHECKING:
    from django.http import HttpRequest, HttpResponse


__all__ = [
    "render_graphiql",
]


def render_graphiql(request: HttpRequest) -> HttpResponse:
    """Render GraphiQL."""
    return render(request, "undine/graphiql.html", context=get_graphiql_context())


class ModuleVersion(StrEnum):
    # Note that changing the versions here will break integrity checks!
    # Integrity values can be generated using the `update_import_map` management command.

    REACT = "19.2.4"
    GRAPHIQL = "5.2.2"
    EXPLORER = "5.1.1"
    GRAPHIQL_REACT = "0.37.3"
    GRAPHIQL_TOOLKIT = "0.11.3"
    GRAPHQL = "16.12.0"


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class ModuleInfoItem:
    latest: str
    """URL to the latest version of the module."""

    version: ModuleVersion
    """Version of the module to use."""

    integrity: str
    """Subresource integrity hash for the module using the given version."""

    standalone: bool = False
    """
    Whether to bundle the module along with all of its `dependencies`,
    excluding `peerDependencies`, into a single JavaScript file.
    """

    external: list[str] = dataclasses.field(default_factory=list)
    """Dependencies that do not need to be imported as they are provided by the host environment."""

    @property
    def url(self) -> str:
        url = self.latest.replace("@latest", f"@{self.version.value}")  # type: ignore[arg-type]
        if self.standalone:
            url += "?standalone"
        if self.external:
            url += f"&external={','.join(self.external)}"
        return url


class ModuleInfo:
    REACT = ModuleInfoItem(
        latest="https://esm.sh/react@latest/",
        version=ModuleVersion.REACT,
        integrity="sha384-w+4QrjFjcEk/AIJMWVUzEOT0QT/s+K0xDQulhBdT/j9PWKIG08nWLFD90h0Yuj5R",
    )
    REACT_DOM = ModuleInfoItem(
        latest="https://esm.sh/react-dom@latest/",
        version=ModuleVersion.REACT,
        integrity="sha384-RWUKf5R/P6U794/65YB+OKpYxJbldLkbnJSNjiByqjgea+hKkDK6vnJ2qut4Lf7W",
    )
    GRAPHIQL = ModuleInfoItem(
        latest="https://esm.sh/graphiql@latest/",
        version=ModuleVersion.GRAPHIQL,
        integrity="sha384-MBVZMq1pmz8DwpwIWPWLk2tmS6tGiSi6WwbXvy9NhuDYASAAWd2m96xbxLqszig9",
    )
    GRAPHIQL_STANDALONE = ModuleInfoItem(
        latest="https://esm.sh/graphiql@latest/",
        version=ModuleVersion.GRAPHIQL,
        integrity="sha384-SzHBEbcQfhvmwqh5Vtat9k7b/kIzmdVO3KMzQiAYwcxCA9x7vZwFRUgjzN1AeV3q",
        standalone=True,
        external=["react", "react-dom", "@graphiql/react", "graphql"],
    )
    EXPLORER = ModuleInfoItem(
        latest="https://esm.sh/@graphiql/plugin-explorer@latest/",
        version=ModuleVersion.EXPLORER,
        integrity="sha384-rR9phbzRkwb/HINixBgg9De/Z/S6G9/OiRX7cVR1AKhP+2AUTfX7wmDT76y5HeSf",
        standalone=True,
        external=["react", "@graphiql/react", "graphql"],
    )
    GRAPHIQL_REACT = ModuleInfoItem(
        latest="https://esm.sh/@graphiql/react@latest/",
        version=ModuleVersion.GRAPHIQL_REACT,
        integrity="sha384-iZsbTy9B0VcX2BOTdqMuX0uJ9Hff5GbG2QeOt4OeMp0GHza76dwQaYQYNYkZkIVq",
        standalone=True,
        external=["react", "react-dom", "graphql", "@graphiql/toolkit", "@emotion/is-prop-valid"],
    )
    GRAPHIQL_TOOLKIT = ModuleInfoItem(
        latest="https://esm.sh/@graphiql/toolkit@latest/",
        version=ModuleVersion.GRAPHIQL_TOOLKIT,
        integrity="sha384-ZsnupyYmzpNjF1Z/81zwi4nV352n4P7vm0JOFKiYnAwVGOf9twnEMnnxmxabMBXe",
        standalone=True,
        external=["graphql"],
    )
    GRAPHQL = ModuleInfoItem(
        latest="https://esm.sh/graphql@latest/",
        version=ModuleVersion.GRAPHQL,
        integrity="sha384-Oosnx71vGzeLRLBj0HblPGTSLgNCn3tEUKVdBubTBLQ9xAW9538VZvjesbp8unrb",
    )
    GRAPHIQL_CSS = ModuleInfoItem(
        latest="https://esm.sh/graphiql@latest/dist/style.css/",
        version=ModuleVersion.GRAPHIQL,
        integrity="sha384-f6GHLfCwoa4MFYUMd3rieGOsIVAte/evKbJhMigNdzUf52U9bV2JQBMQLke0ua+2",
    )
    EXPLORER_CSS = ModuleInfoItem(
        latest="https://esm.sh/@graphiql/plugin-explorer@latest/dist/style.css/",
        version=ModuleVersion.EXPLORER,
        integrity="sha384-vTFGj0krVqwFXLB7kq/VHR0/j2+cCT/B63rge2mULaqnib2OX7DVLUVksTlqvMab",
    )


def get_graphiql_context() -> dict[str, Any]:
    """Get the GraphiQL context."""
    return {
        "http_path": undine_settings.GRAPHQL_PATH,
        "ws_path": undine_settings.WEBSOCKET_PATH,
        "importmap": get_importmap(),
        "graphiql_css": ModuleInfo.GRAPHIQL_CSS.url,
        "explorer_css": ModuleInfo.EXPLORER_CSS.url,
        "graphiql_css_integrity": ModuleInfo.GRAPHIQL_CSS.integrity,
        "explorer_css_integrity": ModuleInfo.EXPLORER_CSS.integrity,
    }


def get_importmap() -> str:
    """Get the importmap for GraphiQL."""
    importmap = {
        "imports": {
            "react": ModuleInfo.REACT.url,
            "react/": ModuleInfo.REACT.url,
            "react-dom": ModuleInfo.REACT_DOM.url,
            "react-dom/": ModuleInfo.REACT_DOM.url,
            "graphiql": ModuleInfo.GRAPHIQL_STANDALONE.url,
            "graphiql/": ModuleInfo.GRAPHIQL.url,
            "@graphiql/plugin-explorer": ModuleInfo.EXPLORER.url,
            "@graphiql/react": ModuleInfo.GRAPHIQL_REACT.url,
            "@graphiql/toolkit": ModuleInfo.GRAPHIQL_TOOLKIT.url,
            "graphql": ModuleInfo.GRAPHQL.url,
            "@emotion/is-prop-valid": "data:text/javascript,",
        },
        "integrity": {
            ModuleInfo.REACT.url: ModuleInfo.REACT.integrity,
            ModuleInfo.REACT_DOM.url: ModuleInfo.REACT_DOM.integrity,
            ModuleInfo.GRAPHIQL.url: ModuleInfo.GRAPHIQL.integrity,
            ModuleInfo.GRAPHIQL_STANDALONE.url: ModuleInfo.GRAPHIQL_STANDALONE.integrity,
            ModuleInfo.EXPLORER.url: ModuleInfo.EXPLORER.integrity,
            ModuleInfo.GRAPHIQL_REACT.url: ModuleInfo.GRAPHIQL_REACT.integrity,
            ModuleInfo.GRAPHIQL_TOOLKIT.url: ModuleInfo.GRAPHIQL_TOOLKIT.integrity,
            ModuleInfo.GRAPHQL.url: ModuleInfo.GRAPHQL.integrity,
        },
    }
    return json.dumps(importmap, indent=2)



from __future__ import annotations

import dataclasses
import inspect
from collections.abc import AsyncGenerator
from contextlib import aclosing, nullcontext
from inspect import isawaitable
from typing import TYPE_CHECKING, Any, Generic

from graphql import GraphQLError, located_error

from undine.exceptions import GraphQLErrorGroup
from undine.optimizer import optimize_async
from undine.typing import TModel
from undine.utils.graphql.utils import pre_evaluate_request_user
from undine.utils.reflection import get_root_and_info_params

if TYPE_CHECKING:
    from collections.abc import AsyncIterable, Callable, Coroutine

    from undine import Entrypoint, GQLInfo, QueryType
    from undine.subscriptions import QueryTypeSignalSubscription

__all__ = [
    "FunctionSubscriptionResolver",
    "SubscriptionValueResolver",
]


@dataclasses.dataclass(frozen=True, slots=True)
class SubscriptionValueResolver:
    """Resolves a value for a subscription."""

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        return root


@dataclasses.dataclass(frozen=True, slots=True)
class FunctionSubscriptionResolver:
    """Subscription resolver for a async generator function or async iterable coroutine."""

    func: Callable[..., AsyncGenerator[Any, None] | Coroutine[Any, Any, AsyncIterable[Any]]]
    entrypoint: Entrypoint

    root_param: str | None = dataclasses.field(default=None, init=False)
    info_param: str | None = dataclasses.field(default=None, init=False)

    def __post_init__(self) -> None:
        params = get_root_and_info_params(self.func)
        object.__setattr__(self, "root_param", params.root_param)
        object.__setattr__(self, "info_param", params.info_param)

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AsyncIterable[Any]:
        return self.subscribe(root, info, **kwargs)

    async def subscribe(self, root: Any, info: GQLInfo, **kwargs: Any) -> AsyncIterable[Any]:
        # Fetch user eagerly so that its available in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        if self.root_param is not None:
            kwargs[self.root_param] = root
        if self.info_param is not None:
            kwargs[self.info_param] = info

        value = self.func(**kwargs)
        iterable = await value if isawaitable(value) else value
        manager = aclosing(iterable) if isinstance(iterable, AsyncGenerator) else nullcontext(iterable)

        async with manager:
            try:
                async for result in iterable:
                    if isinstance(result, GraphQLErrorGroup):
                        yield result.located(path=info.path.as_list())
                        continue

                    if isinstance(result, GraphQLError):
                        yield located_error(result, nodes=info.field_nodes, path=info.path.as_list())
                        continue

                    await self.check_permissions_async(root, info, result)
                    yield result

            except GraphQLErrorGroup as error:
                raise error.located(path=info.path.as_list()) from error

            except Exception as error:
                raise located_error(error, nodes=info.field_nodes, path=info.path.as_list()) from error

    async def check_permissions_async(self, root: Any, info: GQLInfo, instance: dict[str, Any]) -> None:
        if self.entrypoint.permissions_func is not None:
            if inspect.iscoroutinefunction(self.entrypoint.permissions_func):
                await self.entrypoint.permissions_func(root, info, instance)
            else:
                self.entrypoint.permissions_func(root, info, instance)


@dataclasses.dataclass(frozen=True, slots=True)
class ModelSaveSubscriptionResolver(Generic[TModel]):
    """Subscription resolver for a model save signal."""

    subscription: QueryTypeSignalSubscription[TModel]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AsyncIterable[TModel]:
        return self.subscribe(root, info, **kwargs)

    @property
    def query_type(self) -> type[QueryType]:
        return self.subscription.query_type

    async def subscribe(self, root: Any, info: GQLInfo, **kwargs: Any) -> AsyncIterable[TModel]:
        # Fetch user eagerly so that its available in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        query_type = self.subscription.query_type
        subscriber = self.subscription.create_subscriber()
        event_stream = subscriber.subscribe()

        async with aclosing(event_stream):
            try:
                async for event in event_stream:
                    queryset = query_type.__get_queryset__(info)
                    instance = await optimize_async(queryset, info, pk=event.pk)

                    # Instance was deleted before we got a chance to send it to the subscriber.
                    if instance is None:
                        continue

                    await self.check_permissions_async(root, info, instance)
                    yield instance

            except GraphQLErrorGroup as error:
                raise error.located(path=info.path.as_list()) from error

            except Exception as error:
                raise located_error(error, nodes=info.field_nodes, path=info.path.as_list()) from error

    async def check_permissions_async(self, root: Any, info: GQLInfo, instance: TModel) -> None:
        if self.entrypoint.permissions_func is not None:
            if inspect.iscoroutinefunction(self.entrypoint.permissions_func):
                await self.entrypoint.permissions_func(root, info, instance)
            else:
                self.entrypoint.permissions_func(root, info, instance)

        elif inspect.iscoroutinefunction(self.query_type.__permissions__):
            await self.query_type.__permissions__(instance, info)

        else:
            self.query_type.__permissions__(instance, info)


@dataclasses.dataclass(frozen=True, slots=True)
class ModelDeleteSubscriptionResolver(Generic[TModel]):
    """Subscription resolver for a model delete signal."""

    subscription: QueryTypeSignalSubscription[TModel]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AsyncIterable[TModel]:
        return self.subscribe(root, info, **kwargs)

    async def subscribe(self, root: Any, info: GQLInfo, **kwargs: Any) -> AsyncIterable[TModel]:
        # Fetch user eagerly so that its available in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        subscriber = self.subscription.create_subscriber()
        event_stream = subscriber.subscribe()

        async with aclosing(event_stream):
            try:
                async for event in event_stream:
                    await self.check_permissions_async(root, info, event)
                    yield event

            except GraphQLErrorGroup as error:
                raise error.located(path=info.path.as_list()) from error

            except Exception as error:
                raise located_error(error, nodes=info.field_nodes, path=info.path.as_list()) from error

    async def check_permissions_async(self, root: Any, info: GQLInfo, instance: TModel) -> None:
        if self.entrypoint.permissions_func is not None:
            if inspect.iscoroutinefunction(self.entrypoint.permissions_func):
                await self.entrypoint.permissions_func(root, info, instance)
            else:
                self.entrypoint.permissions_func(root, info, instance)



from __future__ import annotations

import io
import logging
from traceback import print_tb
from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    from types import TracebackType

__all__ = [
    "log_traceback",
    "logger",
]


logger = logging.getLogger("undine")


def log_traceback(tb: TracebackType, *, level: Literal[0, 10, 20, 30, 40, 50] = logging.DEBUG) -> None:
    buf = io.StringIO()
    print_tb(tb, file=buf)
    buf.seek(0)
    msg = buf.read()
    logger.log(level, msg)



"""
Contains different types of resolvers for GraphQL operations.
Resolvers must be callables with the following signature:

(root: Root, info: GQLInfo, **kwargs: Any) -> Any
"""

from __future__ import annotations

from .filter import FilterFunctionResolver, FilterModelFieldResolver, FilterQExpressionResolver
from .mutation import (
    BulkCreateResolver,
    BulkDeleteResolver,
    BulkUpdateResolver,
    CreateResolver,
    DeleteResolver,
    UpdateResolver,
)
from .query import (
    ConnectionResolver,
    EntrypointFunctionResolver,
    FieldFunctionResolver,
    GlobalIDResolver,
    InterfaceTypeConnectionResolver,
    InterfaceTypeResolver,
    ModelAttributeResolver,
    ModelGenericForeignKeyResolver,
    ModelManyRelatedFieldResolver,
    ModelSingleRelatedFieldResolver,
    NestedConnectionResolver,
    NestedQueryTypeManyResolver,
    NestedQueryTypeSingleResolver,
    NodeResolver,
    QueryTypeManyResolver,
    QueryTypeSingleResolver,
    UnionTypeConnectionResolver,
    UnionTypeResolver,
)
from .subscription import FunctionSubscriptionResolver, SubscriptionValueResolver

__all__ = [
    "BulkCreateResolver",
    "BulkDeleteResolver",
    "BulkUpdateResolver",
    "ConnectionResolver",
    "CreateResolver",
    "DeleteResolver",
    "EntrypointFunctionResolver",
    "EntrypointFunctionResolver",
    "FieldFunctionResolver",
    "FilterFunctionResolver",
    "FilterModelFieldResolver",
    "FilterQExpressionResolver",
    "FunctionSubscriptionResolver",
    "GlobalIDResolver",
    "InterfaceTypeConnectionResolver",
    "InterfaceTypeResolver",
    "ModelAttributeResolver",
    "ModelGenericForeignKeyResolver",
    "ModelManyRelatedFieldResolver",
    "ModelSingleRelatedFieldResolver",
    "NestedConnectionResolver",
    "NestedQueryTypeManyResolver",
    "NestedQueryTypeSingleResolver",
    "NodeResolver",
    "QueryTypeManyResolver",
    "QueryTypeSingleResolver",
    "SubscriptionValueResolver",
    "UnionTypeConnectionResolver",
    "UnionTypeResolver",
    "UpdateResolver",
]



from __future__ import annotations

import functools
from collections import Counter, defaultdict
from contextlib import contextmanager, suppress
from typing import TYPE_CHECKING, Any, TypeGuard

from django.apps import apps
from django.core.exceptions import FieldDoesNotExist
from django.db import connections, router  # noqa: ICN003
from django.db.models import (
    NOT_PROVIDED,
    CharField,
    F,
    ForeignKey,
    IntegerField,
    ManyToManyField,
    OneToOneField,
    Subquery,
    TextField,
)
from django.db.models.constants import LOOKUP_SEP
from django.db.models.signals import m2m_changed, post_delete, post_save, pre_delete, pre_save
from django.db.utils import DEFAULT_DB_ALIAS, IntegrityError
from django.utils.encoding import force_str

from undine.dataclasses import BulkCreateKwargs
from undine.exceptions import (
    ExpressionMultipleOutputFieldError,
    ExpressionNoOutputFieldError,
    GraphQLDuplicatePrimaryKeysError,
    GraphQLModelConstraintViolationError,
    GraphQLModelNotFoundError,
    GraphQLModelsNotFoundError,
    GraphQLPrimaryKeysMissingError,
    ModelFieldDoesNotExistError,
    ModelFieldNotARelationError,
)
from undine.integrations.modeltranslation import get_translatable_fields, is_translation_field
from undine.settings import undine_settings
from undine.typing import ModelField
from undine.utils.constraints import get_constraint_message

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable

    from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
    from django.core.exceptions import ValidationError
    from django.db.backends.base.features import BaseDatabaseFeatures
    from django.db.models import Field, Manager, ManyToManyRel, Model, QuerySet

    from undine.typing import (
        CombinableExpression,
        GenericField,
        ModelField,
        RelatedField,
        TModel,
        ToManyField,
        ToOneField,
    )

__all__ = [
    "SubqueryCount",
    "convert_integrity_errors",
    "create_union_queryset",
    "determine_output_field",
    "generic_foreign_key_for_generic_relation",
    "generic_relations_for_generic_foreign_key",
    "get_allowed_bulk_create_fields",
    "get_bulk_create_update_fields",
    "get_db_features",
    "get_instance_or_raise",
    "get_instances_or_raise",
    "get_many_to_many_through_field",
    "get_model",
    "get_model_field",
    "get_model_fields_for_graphql",
    "get_pks_from_list_of_dicts",
    "get_related_name",
    "get_save_update_fields",
    "get_validation_error_messages",
    "is_to_many",
    "is_to_one",
    "lookup_to_display_name",
    "set_forward_ids",
]


def get_default_manager(model: type[TModel]) -> Manager[TModel]:
    """Get the default manager for the given model."""
    return model._meta.default_manager  # type: ignore[return-value]


def get_instance_or_raise(*, model: type[TModel], pk: Any) -> TModel:
    """
    Get model instance by the given key with the given primary key.

    :raises GraphQLModelNotFoundError: If an instance for the given primary key does not exists.
    """
    try:
        return get_default_manager(model).get(pk=pk)
    except model.DoesNotExist as error:  # type: ignore[attr-defined]
        raise GraphQLModelNotFoundError(pk=pk, model=model) from error


def get_instances_or_raise(*, model: type[TModel], pks: list[Any]) -> list[TModel]:
    """
    Get model instances by the given primary keys.

    :raises GraphQLModelsNotFoundError: If an instance for any of the given primary keys does not exist.
    """
    instances: list[TModel] = list(get_default_manager(model).filter(pk__in=pks))
    missing = set(pks) - {instance.pk for instance in instances}
    if missing:
        if len(missing) == 1:
            raise GraphQLModelNotFoundError(pk=missing.pop(), model=model)
        raise GraphQLModelsNotFoundError(missing=missing, model=model)
    return instances


def get_pks_from_list_of_dicts(input_data: list[dict[str, Any]]) -> list[Any]:
    """
    Gets primary keys from a list of dicts while validating that each item in list has a pk
    and there are no duplicates. Primary keys are in the same order as in the given list.

    :raises GraphQLPrimaryKeysMissingError: Some items don't have pk.
    :raises GraphQLDuplicatePrimaryKeysError: There are some duplicate primary keys.
    """
    pks = [data["pk"] for data in input_data if "pk" in data]
    if len(pks) != len(input_data):
        raise GraphQLPrimaryKeysMissingError(got=len(pks), expected=len(input_data))

    if len(set(pks)) != len(pks):
        duplicates = [item for item, count in Counter(pks).items() if count > 1]
        raise GraphQLDuplicatePrimaryKeysError(duplicates=duplicates)

    return pks


def generic_relations_for_generic_foreign_key(fk: GenericForeignKey) -> Generator[GenericRelation, None, None]:
    """Get all GenericRelations for the given GenericForeignKey."""
    from django.contrib.contenttypes.fields import GenericRelation  # noqa: PLC0415

    return (field for field in fk.model._meta._relation_tree if isinstance(field, GenericRelation))


def generic_foreign_key_for_generic_relation(relation: GenericRelation) -> GenericForeignKey:
    """Get the GenericForeignKey for the given GenericRelation."""
    from django.contrib.contenttypes.fields import GenericForeignKey  # noqa: PLC0415

    return next(
        field
        for field in relation.related_model._meta.get_fields()  # type: ignore[union-attr]
        if (
            isinstance(field, GenericForeignKey)
            and field.fk_field == relation.object_id_field_name
            and field.ct_field == relation.content_type_field_name
        )
    )


def get_model(*, name: str, app_label: str | None = None) -> type[Model] | None:
    """Get model if it exists in the app. Optionally specify 'app_label' for only look in that app."""
    if app_label is not None:
        try:
            return apps.get_model(app_label, name)
        except LookupError:
            return None

    for app_conf in apps.get_app_configs():
        with suppress(LookupError):
            return app_conf.get_model(name)

    return None


def get_model_field(*, model: type[Model], lookup: str) -> ModelField:
    """
    Gets a model field from the given lookup string.

    :param model: Django model to start finding the field from.
    :param lookup: Lookup string using Django's lookup syntax. E.g. "foo__bar__baz".
    """
    parts = lookup.split(LOOKUP_SEP)
    last = len(parts)
    field: ModelField | None = None

    for part_num, part in enumerate(parts, start=1):
        if part == "pk":
            field = model._meta.pk
        else:
            try:
                field = model._meta.get_field(part)  # type: ignore[assignment]
            except FieldDoesNotExist as error:
                if not part.endswith("_set"):
                    raise ModelFieldDoesNotExistError(field=part, model=model) from error

                # Field might be a reverse many-related field without `related_name`, in which case
                # the `model._meta.fields_map` will store the relation without the "_set" suffix.
                try:
                    field = model._meta.get_field(part.removesuffix("_set"))  # type: ignore[assignment]
                except FieldDoesNotExist as error:
                    raise ModelFieldDoesNotExistError(field=part, model=model) from error

        if part_num == last:
            break

        if not field.is_relation:  # type: ignore[union-attr]
            raise ModelFieldNotARelationError(field=part, model=model)

        model = field.related_model  # type: ignore[union-attr,assignment]

    if field is None:  # pragma: no cover
        raise ModelFieldDoesNotExistError(field=lookup, model=model) from None

    return field


def get_many_to_many_through_field(field: ManyToManyField | ManyToManyRel) -> ForeignKey:
    """
    Get the field on the through model corresponding to the given many to many field.

    Here is a diagram that represents a many-to-many relationship from 'model_1' to 'model_2',
    where 'model_1' has the 'ManyToManyField' to 'model_2'.

    ( model_1 ).<m2m_field> <-- <source>.( through_model ).<target> --> <m2m_rel>.( model_2 )

    For 'model_1.m2m_field', this function returns 'through_model.source'.
    For 'model_2.m2m_rel', this function returns 'through_model.target'.
    """
    if isinstance(field, ManyToManyField):
        return field.remote_field.through._meta.get_field(field.m2m_reverse_field_name())  # type: ignore[union-attr,return-value]
    return field.through._meta.get_field(field.field.m2m_field_name())  # type: ignore[union-attr,return-value]


def get_related_name(related_field: RelatedField | GenericField) -> str:
    """
    Get by which the relation of this field can be used in:

    - Accessing the relation from a model instance: `instance.related_name.all()`
    - Pre-fetching: `qs.select_related("related_name")` or `qs.prefetch_related("related_name")`
    """
    if hasattr(related_field, "accessor_name"):
        return related_field.accessor_name or related_field.name
    if hasattr(related_field, "get_accessor_name"):  # Django < 5.1
        return related_field.get_accessor_name() or related_field.name
    return related_field.name


def get_related_query_name(related_field: RelatedField | GenericField) -> str:
    """
    Name by which the relation of this field can be used in:

    - Query expressions: `qs.filter(query_name__exact=...)`
    """
    if hasattr(related_field, "related_query_name"):
        if callable(related_field.related_query_name):
            return related_field.related_query_name() or related_field.name
        return related_field.related_query_name or related_field.name
    return related_field.name


def get_field_name(related_field: RelatedField | GenericField) -> str:
    """
    Name by which the relation of this field can be used in:

    - Model meta options: `Model._meta.get_fields("query_name")`
    """
    return related_field.name


def get_model_fields_for_graphql(
    model: type[Model],
    *,
    exclude_nonsaveable: bool = False,
) -> Generator[ModelField, None, None]:
    """
    Get all fields from the model that should be included in a GraphQL schema.

    :param model: The model to get fields from.
    :param exclude_nonsaveable: Whether to include non-relational fields that are not editable or not concrete.
    """
    translatable_fields = get_translatable_fields(model)

    include_translatable = undine_settings.MODELTRANSLATION_INCLUDE_TRANSLATABLE
    """Whether to include translatable fields."""

    include_translations = undine_settings.MODELTRANSLATION_INCLUDE_TRANSLATIONS
    """Whether to include translation fields."""

    for model_field in model._meta._get_fields():
        is_relation = bool(getattr(model_field, "is_relation", False))  # Does field reference a relation?
        editable = bool(getattr(model_field, "editable", True))  # Is field value editable by users?
        concrete = bool(getattr(model_field, "concrete", True))  # Does field correspond to a db column?

        if is_relation:
            yield model_field  # type: ignore[misc]
            continue

        if exclude_nonsaveable and (not editable or not concrete):
            continue

        if not include_translatable and model_field.name in translatable_fields:
            continue

        if not include_translations and is_translation_field(model_field):  # type: ignore[arg-type]
            continue

        yield model_field  # type: ignore[misc]


def get_save_update_fields(instance: Model, *fields: str) -> Iterable[str] | None:
    """
    Update fields to use in saving a model instance.

    >>> instance.save(update_fields=get_save_update_fields(instance, *fields))
    """
    # No need to optimize on create.
    if instance.pk is None:
        return None
    unique_fields: set[str] = set(fields) - {"pk", type(instance)._meta.pk.name}
    # Some fields like 'GenericForeignKey' cannot be in the 'update_fields' set.
    # If they are, we cannot optimize the update to only the fields actually updated.
    if unique_fields.issubset(type(instance)._meta._non_pk_concrete_field_names):  # type: ignore[attr-defined]
        return unique_fields or None
    return None


def get_bulk_create_kwargs(model: type[Model], *fields: str) -> BulkCreateKwargs:
    """
    Get arguments to use in 'queryset.bulk_create' for bulk upsert
    where existing instance is determined by its primary key.

    >>> kwargs = get_bulk_create_kwargs(MyModel, *fields)
    >>> MyModel.objects.bulk_create(objs=instances, **kwargs)
    """
    return BulkCreateKwargs(update_fields=get_bulk_create_update_fields(model, *fields))


def get_bulk_create_update_fields(model: type[Model], *fields: str) -> set[str] | None:
    field_map = get_allowed_bulk_create_fields(model=model)
    allowed_fields = {field for field in fields if field in field_map}
    update_fields = set(fields) & allowed_fields  # Remove unallowed fields.
    return update_fields or None


def get_allowed_bulk_create_fields(model: type[Model]) -> dict[str, Field]:
    return {
        field.name: field  # type: ignore[misc]
        for field in model._meta.get_fields()
        if field.concrete and not field.many_to_many and not getattr(field, "primary_key", False)
    }


def is_to_one(field: ModelField | GenericField) -> TypeGuard[ToOneField]:
    is_one_to_one = bool(getattr(field, "one_to_one", False))
    is_many_to_one = bool(getattr(field, "many_to_one", False))
    return is_one_to_one or is_many_to_one


def is_to_many(field: ModelField | GenericField) -> TypeGuard[ToManyField]:
    is_one_to_many = bool(getattr(field, "one_to_many", False))
    is_many_to_many = bool(getattr(field, "many_to_many", False))
    return is_one_to_many or is_many_to_many


def is_generic_foreign_key(field: ModelField | GenericField) -> TypeGuard[GenericForeignKey]:
    is_many_to_one = bool(getattr(field, "many_to_one", False))
    has_content_type_field = hasattr(field, "ct_field")
    has_object_id_field = hasattr(field, "fk_field")
    return is_many_to_one and has_content_type_field and has_object_id_field


def has_default(field: ModelField | GenericField) -> bool:
    has_auto_default = bool(getattr(field, "auto_now", False)) or bool(getattr(field, "auto_now_add", False))
    return has_auto_default or getattr(field, "default", NOT_PROVIDED) is not NOT_PROVIDED


class SubqueryCount(Subquery):
    """
    Count to-many related objects using a subquery.
    Should be used instead of "models.Count" when there might be collisions
    between counted related objects and filter conditions.

    >>> class Foo(Model):
    >>>     number = IntegerField()
    >>>
    >>> class Bar(Model):
    >>>     number = IntegerField()
    >>>     example = ForeignKey(Foo, on_delete=CASCADE, related_name="bars")
    >>>
    >>> foo = Foo.objects.create(number=1)
    >>> Bar.objects.create(example=foo, number=2)
    >>> Bar.objects.create(example=foo, number=2)
    >>>
    >>> foo = (
    >>>     Foo.objects.annotate(count=Count("bars"))
    >>>     .filter(bars__number=2)
    >>>     .first()
    >>> )
    >>> assert foo.count == 2

    This fails and asserts that count is 4. The reason is that Bar objects are
    joined twice: once for the count, and once for the filter. Django does not
    reuse the join, since it is not aware that the join is the same.

    Therefore, do this instead:

    >>> foo = (
    >>>     Foo.objects.annotate(
    >>>         count=SubqueryCount(
    >>>             Bar.objects.filter(example=OuterRef("pk")),
    >>>         ),
    >>>     )
    >>>     .filter(bars__number=2)
    >>>     .first()
    >>> )
    """

    template = "(SELECT COUNT(*) FROM (%(subquery)s) _count)"
    output_field = IntegerField()

    def __repr__(self) -> str:
        try:
            subquery = str(self.query)
        except Exception:  # noqa: BLE001
            subquery = "<subquery>"
        return f"<{self.__class__.__name__}{self.template % {'subquery': subquery}}>"


def determine_output_field(expression: CombinableExpression, *, model: type[Model]) -> Field:
    """Determine the `output_field` for the given expression if it doesn't have one."""
    if hasattr(expression, "output_field"):
        return expression.output_field

    possible_output_fields: dict[type[Field], Any] = {}
    for expr in expression.get_source_expressions():
        if hasattr(expr, "output_field"):
            field: Field = expr.output_field
            possible_output_fields[field.__class__] = field.clone()  # type: ignore[attr-defined]
            continue

        if isinstance(expr, F):
            field = get_model_field(model=model, lookup=expr.name)  # type: ignore[assignment]
            possible_output_fields[field.__class__] = field.clone()  # type: ignore[attr-defined]
            continue

    if len(possible_output_fields) == 0:
        raise ExpressionNoOutputFieldError(expr=expression)

    if len(possible_output_fields) > 1:
        raise ExpressionMultipleOutputFieldError(expr=expression, output_fields=list(possible_output_fields))

    return next(iter(possible_output_fields.values()))


def create_union_queryset(querysets: Iterable[QuerySet]) -> QuerySet:
    """Create a union queryset from the given querysets."""
    # Some databases (e.g. SQLite) don't support slicing or ordering inside the union components.
    # In those cases, we have to remove any ordering from the querysets before union-ing them.
    # This will affect the order of the results, but we can't do anything about that.
    remove_order_by = not all(get_db_features(qs.db).supports_slicing_ordering_in_compound for qs in querysets)
    if remove_order_by:
        return functools.reduce(lambda x, y: x.order_by().union(y.order_by()), querysets)

    return functools.reduce(lambda x, y: x.union(y), querysets)


def get_db_features(db: str = DEFAULT_DB_ALIAS) -> BaseDatabaseFeatures:
    return connections[db].features


def set_forward_ids(instance: Model) -> None:
    """
    Re-set values in related fields for the given instance.

    This prompts their descriptors to update.

    For foreign keys and one-to-one relations, this sets the "_id" value for the field,
    which can be missing if the related instance was set to the field before it was saved.

    For generic foreign keys, this sets the cached content object for the field,
    which would otherwise be missing due to how the GenericForeignKey descriptor works.
    """
    from django.contrib.contenttypes.fields import GenericForeignKey  # noqa: PLC0415

    for field in instance._meta.get_fields():
        if isinstance(field, GenericForeignKey):
            # Note: If the content object was added to the field before it was saved,
            # and is then accessed using the field's descriptor (__get__), this will not work
            # since the content object gets purged from the field cache.
            setattr(instance, field.name, field.get_cached_value(instance, default=None))

        if isinstance(field, ForeignKey | OneToOneField):
            setattr(instance, field.name, getattr(instance, field.name, None))


@contextmanager
def convert_integrity_errors() -> Generator[None, None, None]:
    """Convert IntegrityErrors raised during the context to a GraphQL error."""
    try:
        yield
    except IntegrityError as error:
        msg = get_constraint_message(error.args[0])
        raise GraphQLModelConstraintViolationError(msg) from error


def lookup_to_display_name(lookup: str, field: ModelField) -> str:
    name_conversion_mapping: dict[str, str] = {
        "exact": "",
        "startswith": "starts_with",
        "endswith": "ends_with",
        "isnull": "is_null",
        "isempty": "is_empty",
    }

    # Text fields support case-insensitive lookups, so use those as default lookups instead
    if isinstance(field, CharField | TextField):
        name_conversion_mapping |= {
            "iexact": "",
            "exact": "exact",
            "istartswith": "starts_with",
            "startswith": "starts_with_exact",
            "iendswith": "ends_with",
            "endswith": "ends_with_exact",
            "icontains": "contains",
            "contains": "contains_exact",
        }

    parts: list[str] = []
    for part in lookup.split(LOOKUP_SEP):
        name = name_conversion_mapping.get(part, part)
        if name:
            parts.append(name)

    return "_".join(parts)


def get_validation_error_messages(validation_error: ValidationError) -> dict[str, list[str]]:
    """
    Get dict of error fields to messages inside the given ValidationError.

    >>> get_validation_error_messages(ValidationError("foo"))
    {"": ["foo"]}
    >>> get_validation_error_messages(ValidationError({"foo": "bar"}))
    {"foo": ["bar"]}
    >>> get_validation_error_messages(ValidationError({"foo": ["bar", "baz"]}))
    {"foo": ["bar", "baz"]}
    """
    error_messages: dict[str, list[str]] = defaultdict(list)  # path -> list of messages

    if hasattr(validation_error, "message"):
        message = validation_error.message
        if validation_error.params:
            with suppress(KeyError):  # Allow for mismatching params
                message %= validation_error.params

        error_messages[""].append(force_str(message))
        return error_messages

    if hasattr(validation_error, "error_dict"):
        for field, errors in validation_error.error_dict.items():
            for error in errors:
                sub_messages = get_validation_error_messages(error)
                for path, messages in sub_messages.items():
                    key = f"{field}.{path}" if path else field
                    error_messages[key].extend(messages)

        return error_messages

    for error in validation_error.error_list:
        sub_messages = get_validation_error_messages(error)
        for path, messages in sub_messages.items():
            error_messages[path].extend(messages)

    return error_messages


@contextmanager
def use_save_signals(
    model: type[Model],
    instances: Iterable[Model],
    update_fields: set[str] | None,
) -> Generator[None, None, None]:
    if pre_save.has_listeners(model):
        for instance in instances:
            pre_save.send(
                sender=model,
                instance=instance,
                raw=False,
                using=router.db_for_write(model, instance=instance),
                update_fields=list(update_fields or []),
            )

    yield

    if post_save.has_listeners(model):
        for instance in instances:
            post_save.send(
                sender=model,
                instance=instance,
                created=True,
                update_fields=list(update_fields or []),
                raw=False,
                using=router.db_for_write(model, instance=instance),
            )


@contextmanager
def use_delete_signals(
    model: type[Model],
    instances: Iterable[Model],
) -> Generator[None, None, None]:
    if pre_delete.has_listeners(model):
        for instance in instances:
            pre_delete.send(
                sender=model,
                instance=instance,
                using=router.db_for_write(model, instance=instance),
                origin=instance,
            )

    yield

    if post_delete.has_listeners(model):
        for instance in instances:
            post_delete.send(
                sender=model,
                instance=instance,
                using=router.db_for_write(model, instance=instance),
                origin=instance,
            )


@contextmanager
def use_m2m_remove_signals(
    model: type[Model],
    source_to_removed_target_pks: dict[Model, set[Any]],
    *,
    target_name: str,
    reverse: bool,
) -> Generator[None, None, None]:
    if not m2m_changed.has_listeners(model):
        yield
        return

    target_model = getattr(model, target_name).field.remote_field.model

    for source, pk_set in source_to_removed_target_pks.items():
        m2m_changed.send(
            sender=model,
            action="pre_remove",
            instance=source,
            reverse=reverse,
            model=target_model,
            pk_set=pk_set,
            using=router.db_for_write(model, instance=source),
        )

    yield

    for source, pk_set in source_to_removed_target_pks.items():
        m2m_changed.send(
            sender=model,
            action="post_remove",
            instance=source,
            reverse=reverse,
            model=target_model,
            pk_set=pk_set,
            using=router.db_for_write(model, instance=source),
        )


@contextmanager
def use_m2m_add_signals(
    model: type[Model],
    source_to_added_target_pks: dict[Model, set[Any]],
    *,
    target_name: str,
    reverse: bool,
) -> Generator[None, None, None]:
    if not m2m_changed.has_listeners(model):
        yield
        return

    target_model = getattr(model, target_name).field.remote_field.model

    for source, pk_set in source_to_added_target_pks.items():
        m2m_changed.send(
            sender=model,
            action="pre_add",
            instance=source,
            reverse=reverse,
            model=target_model,
            pk_set=pk_set,
            using=router.db_for_write(model, instance=source),
        )

    yield

    for source, pk_set in source_to_added_target_pks.items():
        m2m_changed.send(
            sender=model,
            action="post_add",
            instance=source,
            reverse=reverse,
            model=target_model,
            pk_set=pk_set,
            using=router.db_for_write(model, instance=source),
        )



from __future__ import annotations

import base64

from graphql import GraphQLBoolean, GraphQLField, GraphQLID, GraphQLNonNull, GraphQLString
from graphql.type.scalars import serialize_id

from undine import InterfaceField, InterfaceType, QueryType, UnionType
from undine.pagination import PaginationHandler
from undine.settings import undine_settings
from undine.utils.graphql.type_registry import get_or_create_graphql_object_type
from undine.utils.reflection import is_subclass

__all__ = [
    "Connection",
    "Node",
    "PageInfoType",
    "cursor_to_offset",
    "decode_base64",
    "encode_base64",
    "from_global_id",
    "offset_to_cursor",
    "to_global_id",
]


class Node(InterfaceType):
    """An interface for objects with Global IDs."""

    id = InterfaceField(GraphQLNonNull(GraphQLID), description="The Global ID of an object.", field_name="pk")


class Connection:
    """A wrapper for paginating a `QueryType` using Relay Connections."""

    def __init__(
        self,
        ref: type[QueryType | UnionType | InterfaceType],
        /,
        *,
        page_size: int | None = undine_settings.PAGINATION_PAGE_SIZE,
        pagination_handler: type[PaginationHandler] = PaginationHandler,
        description: str | None = None,
    ) -> None:
        """
        Create a new Connection.

        :param ref: The `QueryType`, `UnionType`, or `InterfaceType` to use.
        :param page_size: Maximum number of items to return in a page. No limit if `None`.
        :param pagination_handler: Handler to use for pagination.
        :param description: Description for the created GraphQL type.
        """
        self.query_type = ref if is_subclass(ref, QueryType) else None
        self.union_type = ref if is_subclass(ref, UnionType) else None
        self.interface_type = ref if is_subclass(ref, InterfaceType) else None

        self.page_size = page_size
        self.pagination_handler = pagination_handler
        self.description = description


def encode_base64(string: str) -> str:
    return base64.b64encode(string.encode("utf-8")).decode("ascii")


def decode_base64(string: str) -> str:
    return base64.b64decode(string.encode("ascii")).decode("utf-8")


def offset_to_cursor(typename: str, offset: int) -> str:
    """Create the cursor string from an offset."""
    return encode_base64(f"connection:{typename}:{offset}")


def cursor_to_offset(typename: str, cursor: str) -> int:
    """Extract the offset from the cursor string."""
    return int(decode_base64(cursor).removeprefix(f"connection:{typename}:"))


def to_global_id(typename: str, object_id: str | int) -> str:
    """
    Takes a typename and an object ID specific to that type,
    and returns a "Global ID" that is unique among all types.
    """
    return encode_base64(f"ID:{typename}:{serialize_id(object_id)}")


def from_global_id(global_id: str) -> tuple[str, str | int]:
    """
    Takes the "Global ID" created by `to_global_id`,
    and returns the typename and object ID used to create it.
    """
    global_id = decode_base64(global_id)
    _, typename, object_id = global_id.split(":")
    if object_id.isdigit():
        return typename, int(object_id)
    return typename, object_id


PageInfoType = get_or_create_graphql_object_type(
    name="PageInfo",
    description="Information about the current state of the pagination.",
    fields={
        "hasNextPage": GraphQLField(
            GraphQLNonNull(GraphQLBoolean),
            description="Are there more items after the current page?",
        ),
        "hasPreviousPage": GraphQLField(
            GraphQLNonNull(GraphQLBoolean),
            description="Are there more items before the current page?",
        ),
        "startCursor": GraphQLField(
            GraphQLString,  # null if no results
            description=(
                "Value of the first cursor in the current page. "
                "Use as the value for the `before` argument to paginate backwards."
            ),
        ),
        "endCursor": GraphQLField(
            GraphQLString,  # null if no results
            description=(
                "Value of the last cursor in the current page. "
                "Use as the value for the `after` argument to paginate forwards."
            ),
        ),
    },
)



from __future__ import annotations

import datetime
import zoneinfo
from inspect import cleandoc

from django.conf import settings
from django.utils import dateparse
from django.utils.timezone import get_default_timezone, is_naive

from undine.exceptions import GraphQLScalarInvalidValueError

from ._definition import ScalarType

__all__ = [
    "GraphQLDateTime",
    "datetime_scalar",
]


datetime_scalar: ScalarType[datetime.datetime, str] = ScalarType(
    name="DateTime",
    description=cleandoc(
        """
        Represents a date and time value as specified by ISO 8601.
        Maps to the Python `datetime.datetime` type.
        """
    ),
    specified_by_url="https://datatracker.ietf.org/doc/html/rfc3339#section-5.6",
)

GraphQLDateTime = datetime_scalar.as_graphql_scalar()


@datetime_scalar.parse.register
def _(value: str) -> datetime.datetime:
    parsed_value: datetime.datetime | None = dateparse.parse_datetime(value)
    if parsed_value is None:
        raise GraphQLScalarInvalidValueError(typename="datetime")

    if settings.USE_TZ and is_naive(parsed_value):
        parsed_value = parsed_value.replace(tzinfo=zoneinfo.ZoneInfo("UTC"))
        parsed_value = parsed_value.astimezone(get_default_timezone())

    return parsed_value


@datetime_scalar.serialize.register
def _(value: str) -> str:
    parsed_value: datetime.datetime | None = dateparse.parse_datetime(value)
    if parsed_value is None:
        raise GraphQLScalarInvalidValueError(typename="datetime")

    if settings.USE_TZ and is_naive(parsed_value):
        parsed_value = parsed_value.replace(tzinfo=zoneinfo.ZoneInfo("UTC"))
        parsed_value = parsed_value.astimezone(get_default_timezone())

    return parsed_value.isoformat()


@datetime_scalar.serialize.register
def _(value: datetime.datetime) -> str:
    return value.isoformat()



from __future__ import annotations

from django.core.validators import validate_ipv6_address

from ._definition import ScalarType

__all__ = [
    "GraphQLIPv6",
    "ipv6_scalar",
]


ipv6_scalar: ScalarType[str, str] = ScalarType(
    name="IPv6",
    description="Represents a valid IPv6 address.",
    specified_by_url="https://datatracker.ietf.org/doc/html/rfc8200",
)

GraphQLIPv6 = ipv6_scalar.as_graphql_scalar()


@ipv6_scalar.serialize.register
@ipv6_scalar.parse.register
def _(value: str) -> str:
    validate_ipv6_address(value)
    return value



from __future__ import annotations

from inspect import cleandoc

from ._definition import ScalarType

__all__ = [
    "GraphQLNull",
    "null_scalar",
]


null_scalar: ScalarType[None, None] = ScalarType(
    name="Null",
    description=cleandoc(
        """
        Represents represents an always null value.
        Maps to the Python `None` value.
        """
    ),
)

GraphQLNull = null_scalar.as_graphql_scalar()


@null_scalar.serialize.register
@null_scalar.parse.register
def _(value: None) -> None:
    return value



from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any

from graphql import (
    DirectiveLocation,
    GraphQLEnumType,
    GraphQLInputObjectType,
    GraphQLInterfaceType,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLSchema,
    GraphQLUnionType,
    validate_schema,
)

from undine.exceptions import UndineErrorGroup
from undine.settings import undine_settings
from undine.utils.graphql.type_registry import get_registered_directives
from undine.utils.graphql.utils import check_directives
from undine.utils.logging import logger
from undine.utils.reflection import get_signature

if TYPE_CHECKING:
    from graphql import GraphQLNamedType

    from undine import RootType
    from undine.directives import Directive

__all__ = [
    "create_schema",
]


def create_schema(
    *,
    query: type[RootType],
    mutation: type[RootType] | None = None,
    subscription: type[RootType] | None = None,
    description: str | None = None,
    schema_definition_directives: list[Directive] | None = None,
    extensions: dict[str, Any] | None = None,
) -> GraphQLSchema:
    """
    Creates the GraphQL schema.

    :param query: The `RootType` for the `Query` operations.
    :param mutation: The `RootType` for the `Mutation` operations.
    :param subscription: The `RootType` for the `Subscription` operations.
    :param description: The description for the schema.
    :param schema_definition_directives: The directives to add to the schema definition.
    :param extensions: The extensions for the schema.
    """
    started = time.perf_counter()
    extensions = extensions or {}

    if schema_definition_directives is not None:
        check_directives(schema_definition_directives, location=DirectiveLocation.SCHEMA)
        extensions[undine_settings.SCHEMA_DIRECTIVES_EXTENSIONS_KEY] = schema_definition_directives

    directives = get_registered_directives()

    logger.debug("Creating Query type...")
    query_object_type: GraphQLObjectType = query.__output_type__()

    mutation_object_type: GraphQLObjectType | None = None
    if mutation is not None:
        logger.debug("Creating Mutation type...")
        mutation_object_type = mutation.__output_type__()

    subscription_object_type: GraphQLObjectType | None = None
    if subscription is not None:
        logger.debug("Creating Subscription type...")
        subscription_object_type = subscription.__output_type__()

    logger.debug("Creating GraphQL schema...")

    schema = GraphQLSchema(
        query=query_object_type,
        mutation=mutation_object_type,
        subscription=subscription_object_type,
        directives=directives,
        description=description,
        extensions=extensions,
    )

    sort_schema_types(schema)

    logger.debug("Validating GraphQL schema...")

    schema_validation_errors = validate_schema(schema)
    if schema_validation_errors:
        msg = "Schema validation failed"
        raise UndineErrorGroup(schema_validation_errors, msg=msg)

    elapsed = time.perf_counter() - started
    logger.debug(f"GraphQL schema created successfully in {elapsed}s!")

    # Clear cached signatures for functions to reduce memory usage.
    get_signature.cache.clear()
    return schema


def sort_schema_types(schema: GraphQLSchema) -> None:
    """Sort Schema types by type and name so that browsing GraphiQL is easier."""

    def key_func(item: tuple[str, GraphQLNamedType]) -> tuple[int, str]:
        match item[1]:
            # Put RootTypes at the end.
            case schema.query_type:
                type_order = 8
            case schema.mutation_type:
                type_order = 9
            case schema.subscription_type:
                type_order = 10
            # Sort more generic types first and then more specific types.
            case GraphQLScalarType():
                type_order = 1
            case GraphQLEnumType():
                type_order = 2
            case GraphQLInterfaceType():
                type_order = 3
            case GraphQLUnionType():
                type_order = 4
            case GraphQLObjectType():
                type_order = 5
            case GraphQLInputObjectType():
                type_order = 6
            case _:
                type_order = 7

        return type_order, item[0]

    schema.type_map = dict(sorted(schema.type_map.items(), key=key_func))



from __future__ import annotations

from django.urls import path

from undine.settings import undine_settings

from .views import persisted_documents_view

app_name = "undine.persisted_documents"

urlpatterns = [
    path(
        undine_settings.PERSISTED_DOCUMENTS_PATH,
        persisted_documents_view,
        name=undine_settings.PERSISTED_DOCUMENTS_VIEW_NAME,
    ),
]
